import numpy as np
from autograd import grad as ag_grad
from autograd import jacobian as ag_jacobian
from autograd import hessian as ag_hessian
import warnings
from typing import List
from .abstract_function import AbstractFunction
from .utilities import check_input


class StochasticFunction(AbstractFunction):
    """Stochastic function

    .. math::

        f(x)=\\mathbb{E}[h(x)]

    with :math:`x: \\mathbb{R}^{n}`.

    The :class:`random_batch` method extract a batch from the function and :class:`batch_subgradient`, :class:`batch_jacobian` and :class:`batch_hessian` methods return a subgradient, jacobian and hessian computed at on the last batch.

    Args:
        fn_list (list): list of AbstractFunction objects
        probabilities (list, optional): list with the probabilities of drawing each function. Default is None which leads to uniform probabilities

    Raises:
        TypeError: fn_list input must be a list of functions
        ValueError: All functions must have the same input/output shape
        TypeError: probabilities argument must be a list of floats
        ValueError: inputs must have the same lenght
        ValueError: provided probabilities must sum to 1
        NotImplementedError: only 1, 2 and inf norms are currently supported
    """

    def __init__(self, fn_list: List[AbstractFunction], probabilities: List[float]):
        if not isinstance(fn_list, list):
            raise TypeError("fn_list input must be a list of functions")
        for f in fn_list:
            if not isinstance(fn, AbstractFunction):
                raise TypeError("fn_list input must be a list of functions")

        input_shape = fn_list[0].input_shape
        output_shape = fn_list[0].output_shape
        self.differentiable = True
        self.affine = True
        self.quadratic = True
        for item in fn_list:
            if item.input_shape != input_shape or item.output_shape != output_shape:
                raise ValueError("All functions must have the same input/output shape")
            if not item.is_differentiable:
                self.differentiable = False
            if not item.is_affine:
                self.affine = False
            if not item.is_quadratic:
                self.quadratic = False

        self.fn_list = fn_list
        self.items = len(fn_list)
        self.input_shape = input_shape
        self.output_shape = output_shape

        if probabilities is not None:
            if not isinstance(probabilities, list):
                raise TypeError("probabilities argument must be a list")
            for p in probabilities:
                if not isinstance(p, float):
                    raise TypeError("probabilities argument must be a list")
            if len(probabilities) != len(fn_list):
                raise ValueError("inputs must have the same lenght")
            if sum(probabilities) != 1.0:
                raise ValueError("provided probabilities must sum to 1")
        else:
            probabilities = (np.ones(self.items)/self.items).tolist()
        self.probabilities = probabilities

        super().__init__()

    @check_input
    def eval(self, x: np.ndarray) -> np.ndarray:
        value = 0
        for item in range(self.items):
            value += self.probabilities[item] * self.fn_list[item].eval(x)
        return value.reshape(self.output_shape)

    def random_batch(self, batch_size: int = 1):
        """generate a random batch from the function.
        
        Args:
            batch_size (int, optional): batch size. Defaults to 1.
        """
        self.selected = np.random.choice(self.items,
                                         size=batch_size,
                                         p=self.probabilities)
        self._setbatchdiff()

    def _setbatchdiff(self):
        self._batch_subgradient = ag_grad(self._batch_eval)
        self._batch_jacobian = ag_jacobian(self._batch_eval)
        self._batch_hessian = ag_hessian(self._batch_eval)
    
    def _batch_eval(self, x: np.ndarray) -> np.ndarray:
        value = 0
        for sample in self.selected:
            value += self.fn_list[sample].eval(x)
        return value.reshape(self.output_shape)

    @check_input
    def batch_jacobian(self, x: np.ndarray):
        """evaluate the jacobian on the current batch
        
        Args:
            x (np.ndarray): point
        Returns:
            numpy.ndarray: jacobian
        """
        return np.squeeze(self._batch_jacobian(x))
    
    @check_input
    def batch_hessian(self, x: np.ndarray):
        """evaluate the hessian on the current batch
        
        Args:
            x (np.ndarray): point
        
        Returns:
            numpy.ndarray: hessian
        """
        return np.squeeze(self._batch_hessian(x))

    @check_input
    def batch_subgradient(self, x: np.ndarray):
        """evaluate the subgradient on the current batch
        
        Args:
            x (np.ndarray): point
        
        Raises:
            ValueError: Only functions with scalar output have a subgradient
        
        Returns:
            numpy.ndarray: subgradient
        """
        if not self.output_shape == (1, 1):
            raise ValueError("Undefined subgradient for output_shape={}".format(output_shape))
        else:
            return self._batch_subgradient(x).reshape(self.input_shape)


    # @check_input
    # def jacobian(self, x: np.ndarray, batch_size: int = 1, **kwargs) -> np.ndarray:
    #     selected = np.random.choice(self.items,
    #                                 size=batch_size,
    #                                 p=self.probabilities)

    #     jac = self.fn_list[selected[0]].jacobian(x)
    #     if selected.size > 1:
    #         for sel in selected[1:]:
    #             jac += self.fn_list[sel].jacobian(x)

    #     return jac



from .problem import Problem
from .constraint_coupled_problem import ConstraintCoupledProblem, ConstraintCoupledMILP
from .linear_problem import LinearProblem
from .milp import MixedIntegerLinearProblem
from .conv_milp import ConvexifiedMILP
from .quadratic_problem import QuadraticProblem
from .projection_problem import ProjectionProblem


import numpy as np
import time
from copy import deepcopy
from typing import Union, Callable
from ..agents import Agent
from ..constraints import AbstractSet, Constraint
from ..problems import ProjectionProblem
from .consensus import Consensus, AsynchronousConsensus


class SetMembership(Consensus):
    """Distributed Set Membership Algorithm [FaGa18]_

    From the perspective of agent :math:`i` the algorithm works as follows. For :math:`k=0,1,\\dots`

    .. math::

        X_i^{k+1} &= X_i^k \cap M_i^{k+1}

        z_i^{k} &= \sum_{j=1}^N w_{ij} x_j^k

        x_i^{k+1} &= \Pi_{X_i^{k+1}}[z_i^k]

    where :math:`x_i,z_i\\in\\mathbb{R}^n`, :math:`M_i^{k+1},X_i^{k+1}\\subseteq\mathbb{R}^n` are the current feasible (measurement) set and the feasible (parameter) set respectively, and :math:`\\Pi_X[]` denotes the projection operator over the set :math:`X`.


    Args:
        agent (Agent): agent to execute the algorithm (must be a Agent)
        initial_condition (numpy.ndarray): initial condition
        enable_log (bool): True for enabling log

    Attributes:
        agent (Agent): agent to execute the algorithm
        x0 (numpy.ndarray): initial condition
        x (numpy.ndarray): current value of the local solution
        shape (tuple): shape of the variable
        x_neigh (dict): dictionary containing the local solution of the (in-)neighbors
        enable_log (bool): True for enabling log
    """

    def __init__(self, agent: Agent, initial_condition: np.ndarray, enable_log=False):
        if not isinstance(agent, Agent):
            raise TypeError("Must provide an Agent as agent")

        super(SetMembership, self).__init__(agent, initial_condition, enable_log)

        self.parameter_set = None
        self.measurement_set = None
        self.measure_generator = None 
    
    def set_measure_generator(self, generator: Callable):
        """set the measure generator

        Args:
            generator: measure generator
        """
        self.measure_generator = generator

    def measure(self):
        """Takes a new measurement and updates parameter_set
        """
        new_set = self.measure_generator()
        self.measurement_set = deepcopy(new_set)

        if self.parameter_set is not None:
            if isinstance(self.parameter_set, AbstractSet):
                try:
                    self.parameter_set.intersection(new_set)
                except ValueError:
                    if isinstance(self.measurement_set, AbstractSet):
                        self.parameter_set = self.parameter_set.to_constraints() + new_set.to_constraints()
                    else:
                        if isinstance(self.measurement_set, Constraint):
                            self.parameter_set = self.parameter_set.to_constraints().append(new_set)
                        elif isinstance(self.measurement_set, list):
                            self.parameter_set = self.parameter_set.to_constraints().extend(new_set)
            elif isinstance(self.parameter_set, list):
                if isinstance(self.measurement_set, Constraint):
                    self.parameter_set = self.parameter_set.to_constraints().append(new_set)
                elif isinstance(self.measurement_set, list):
                    self.parameter_set = self.parameter_set.to_constraints().extend(new_set)
            elif isinstance(self.parameter_set, Constraint):
                if isinstance(self.measurement_set, Constraint):
                    self.parameter_set = [self.parameter_set].append(new_set)
                elif isinstance(self.measurement_set, list):
                    self.parameter_set = [self.parameter_set].extend(new_set)
        else:
            if isinstance(self.measurement_set, AbstractSet):
                self.parameter_set = new_set
            if isinstance(self.measurement_set, Constraint):
                self.parameter_set = [new_set]
            elif isinstance(self.measurement_set, list):
                self.parameter_set = new_set

    def _update_local_solution(self, x_new: np.ndarray):
        self.measure()
        if isinstance(self.parameter_set, AbstractSet):
            self.x = self.parameter_set.projection(x_new)
        else:  
            constraints = deepcopy(self.parameter_set)
            pb = ProjectionProblem(constraints, x_new)
            self.x = pb.solve()



class AsynchronousSetMembership(AsynchronousConsensus):
    """Asynchronous Distributed Set Membership Algorithm [FaGa19]_ 

    Args:
        agent (Agent): agent to execute the algorithm (must be a Agent)
        initial_condition (numpy.ndarray): initial condition
        enable_log (bool): True for enabling log

    Attributes:
        agent (Agent): agent to execute the algorithm
        x0 (numpy.ndarray): initial condition
        x (numpy.ndarray): current value of the local solution
        shape (tuple): shape of the variable
        x_neigh (dict): dictionary containing the local solution of the (in-)neighbors
        enable_log (bool): True for enabling log
        timestamp_sequence (list): list of timestamps
    """

    def __init__(self, agent: Agent, initial_condition: np.ndarray, **kwargs):
        if not isinstance(agent, Agent):
            raise TypeError("Must provide an Agent as agent")

        super(AsynchronousSetMembership, self).__init__(agent,
                                                        initial_condition, 
                                                        **kwargs)
        self.parameter_set = None
        self.measurement_set = None
        self.measure_generator = None 
    
    def set_measure_generator(self, generator: Callable):
        """set the measure generator

        Args:
            generator: measure generator
        """
        self.measure_generator = generator

    def measure(self):
        """Takes a new measurement and updates parameter_set
        """
        new_set = self.measure_generator()
        self.measurement_set = deepcopy(new_set)

        if self.parameter_set is not None:
            if isinstance(self.parameter_set, AbstractSet):
                try:
                    self.parameter_set.intersection(new_set)
                except ValueError:
                    if isinstance(self.measurement_set, AbstractSet):
                        self.parameter_set = self.parameter_set.to_constraints() + new_set.to_constraints()
                    else:
                        if isinstance(self.measurement_set, Constraint):
                            self.parameter_set = self.parameter_set.to_constraints().append(new_set)
                        elif isinstance(self.measurement_set, list):
                            self.parameter_set = self.parameter_set.to_constraints().extend(new_set)
            elif isinstance(self.parameter_set, list):
                if isinstance(self.measurement_set, Constraint):
                    self.parameter_set = self.parameter_set.to_constraints().append(new_set)
                elif isinstance(self.measurement_set, list):
                    self.parameter_set = self.parameter_set.to_constraints().extend(new_set)
            elif isinstance(self.parameter_set, Constraint):
                if isinstance(self.measurement_set, Constraint):
                    self.parameter_set = [self.parameter_set].append(new_set)
                elif isinstance(self.measurement_set, list):
                    self.parameter_set = [self.parameter_set].extend(new_set)
        else:
            if isinstance(self.measurement_set, AbstractSet):
                self.parameter_set = new_set
            if isinstance(self.measurement_set, Constraint):
                self.parameter_set = [new_set]
            elif isinstance(self.measurement_set, list):
                self.parameter_set = new_set

    def _update_local_solution(self, x_new: np.ndarray):
        self.measure()
        if isinstance(self.parameter_set, AbstractSet):
            self.x = self.parameter_set.projection(x_new)
        else:  
            constraints = deepcopy(self.parameter_set)
            pb = ProjectionProblem(constraints, x_new)
            self.x = pb.solve()




import numpy as np
import warnings
from .abstract_function import AbstractFunction
from .affine_form import AffineForm
from .utilities import check_input


class Variable(AffineForm):
    """Variable, basic function

    .. math::

        f(x) = x 

    with :math:`x\\in \\mathbb{R}^{n}`

    Args:
        n (int): dimension of the decision variable: (n,1)

    Raises:
        TypeError: input dimension must be an int
    """

    def __init__(self, n: int):
        if not isinstance(n, int):
            raise TypeError("Input must be an int")
        self.input_shape = (n, 1)
        self.output_shape = (n, 1)

        # affine parameters
        self.A = np.eye(self.input_shape[0])
        self.b = np.zeros(self.output_shape)
        self.fn = self

        self.differentiable = True
        self.affine = True
        super(AffineForm, self).__init__()

        import cvxpy as cvx
        self.cvx_var = cvx.Variable(self.input_shape[0], var_id=0)

    def _expression(self):
        expression = self.__class__.__name__+'{}'.format(self.input_shape)
        return expression
    
    def _to_cvxpy(self):
        return self.cvx_var

    @check_input
    def eval(self, x: np.ndarray) -> np.ndarray:
        return x.reshape(self.output_shape)

    def _extend_variable(self, n_var, axis, pos):
        if axis != 0:
            raise NotImplementedError("Only axis=0 is supported")
        n = self.input_shape[0]
        A = np.zeros((n + n_var, n))
        A[pos:pos+n, :] = np.eye(n)
        return A @ Variable(n + n_var)


import numpy as np


class AbstractSet:
    """Abstract constraint set
    """

    def projection(self, x):
        """Project a point onto the set
        """
        pass

    def to_constraints(self):
        """Convert the set in a list of Constraints
        """
        pass


class Box(AbstractSet):
    """Box set

    :math:`X = \\{x\\mid l\\leq x \\leq u\\}`
    with :math:`l,x,u\\in \\mathbb{R}^{n}`

    Args:
        lower_bound (numpy.ndarray, optional): array with lower bounds for each axis. Defaults to None.
        upper_bound (numpy.ndarray, optional): array with upper bounds for each axis. Defaults to None.

    Attributes:
        lower_bound (numpy.ndarray): lower bounds
        upper_bound (numpy.ndarray): upper bounds
        input_shape (tuple): space dimensions
    """

    def __init__(self, lower_bound=None, upper_bound=None):
        if lower_bound is None and upper_bound is None:
            raise ValueError("At least one argument must be provided.")

        if (lower_bound is not None):
            if not isinstance(lower_bound, np.ndarray):
                raise ValueError("lower_bound must be numpy.ndarray")
            elif len(lower_bound.shape) > 2 or lower_bound.shape[1] != 1:
                raise ValueError(
                    "inputs must be 2D numpy.ndarray with shape (n,1), n>=1")
            self.lower_bound = lower_bound

        if (upper_bound is not None):
            if not isinstance(upper_bound, np.ndarray):
                raise ValueError("upper_bound must be numpy.ndarray")
            elif len(upper_bound.shape) > 2 or upper_bound.shape[1] != 1:
                raise ValueError(
                    "inputs must be 2D numpy.ndarray with shape (n,1), n>=1")
            self.upper_bound = upper_bound

        if (lower_bound is not None) and (upper_bound is not None):
            if lower_bound.shape != upper_bound.shape:
                raise ValueError(
                    "Lower and upper bounds must have the same shape")
            else:
                self.input_shape = lower_bound.shape
        else:
            if lower_bound is None:
                self.lower_bound = -np.inf * np.ones(upper_bound.shape)
                self.input_shape = lower_bound.shape
            elif upper_bound is None:
                self.upper_bound = np.inf * np.ones(lower_bound.shape)
                self.input_shape = lower_bound.shape

    def intersection(self, box):
        """Compute the intersection with another box

        Args:
            box (Box): box to compute the intersection with

        Raises:
            ValueError: Only intersection with another box is supported
            ValueError: The two boxex must have the same input_shape
        """
        if not isinstance(box, Box):
            raise ValueError(
                "Only intersection with another box is supported")
        if self.input_shape != box.input_shape:
            raise ValueError(
                "Boxes must have the same shape {}.".format(self.input_shape))
        else:
            self.upper_bound = np.min([self.upper_bound, box.upper_bound], axis=0)
            self.lower_bound = np.max([self.lower_bound, box.lower_bound], axis=0)

    def projection(self, x):
        """Project a point onto the box

        Arguments:
            x (numpy.ndarray): Point to be projected

        Returns:
            numpy.ndarray: projected point
        """
        if not isinstance(x, np.ndarray):
            raise ValueError("input must be a numpy.ndarray")
        if x.shape != self.input_shape:
            raise ValueError(
                "Dimensions mismatch. Required input shape is {}".format(self.input_shape))
        x_projected = np.zeros(self.input_shape)
        x_projected += x
        for i in range(self.input_shape[0]):
            if x[i] < self.lower_bound[i]:
                x_projected[i] = self.lower_bound[i]
            elif x[i] > self.upper_bound[i]:
                x_projected[i] = self.upper_bound[i]
        return x_projected

    def to_constraints(self):
        from ..functions import Variable
        x = Variable(self.input_shape[0])
        upper_bound = x <= self.upper_bound
        lower_bound = x >= self.lower_bound
        return [upper_bound, lower_bound]


class Strip(AbstractSet):
    """Strip set

    :math:`X = \\{x\\mid -w<=|a^\\top x - s|<=w\\}`
    with :math:`a,x\\in \\mathbb{R}^{n}` and :math:`w\\in\\mathbb{R}`

    Args:
        regressor (numpy.ndarray): regressor of the strip
        shift (float): shift from the origin
        width (float): width of the strip

    Attributes:
        regressor (numpy.ndarray): regressor of the strip
        shift (float): shift from the origin
        upper (float): upper border (shift + half width)
        lower (float): lower border (shift - half width)
        input_shape (tuple): space dimensions
    """

    def __init__(self, regressor, shift, width):
        if not isinstance(regressor, np.ndarray):
            raise ValueError("regressor must be numpy.ndarray")
        if len(regressor.shape) > 2 or regressor.shape[1] != 1:
            raise ValueError(
                "regressor must be 2D numpy.ndarray with shape (n,1), n>=1")
        self.regressor = regressor
        self.shift = shift
        self.upper = shift + width/2.0
        self.lower = shift - width/2.0
        self.input_shape = regressor.shape

    def __str__(self):
        description = r'{}<=|{}^T x - {}|<={}'.format(
            self.lower, self.regressor, self.shift, self.upper)
        return description

    def intersection(self, strip):
        """Compute the intersection with another strip

        Args:
            strip (Strip): strip to compute the intersection with

        Raises:
            ValueError: Only intersection with another strip is supported
            ValueError: The two strips must have the same regressor
        """
        if not isinstance(strip, Strip):
            raise ValueError(
                "Only intersection with another strip is supported")
        if self.regressor.any() != strip.regressor.any():
            raise ValueError(
                "Strips must have the same regressor {}.".format(self.regressor))
        else:
            self.upper = np.min([self.upper, strip.upper])
            self.lower = np.max([self.lower, strip.lower])

    def projection(self, x):
        """Project a point onto the strip

        Arguments:
            x (numpy.ndarray): Point to be projected

        Returns:
            numpy.ndarray: projected point
        """
        if self.input_shape != x.shape:
            raise ValueError("shape of the point {} is not equal to the shape of the regressor {}".format(
                x.shape, self.regressor.shape))
        distance = self.regressor.T.dot(x)
        if distance > self.upper:
            x_projected = x + float(self.upper - distance) * self.regressor
            return x_projected
        elif distance < self.lower:
            x_projected = x + float(self.lower - distance) * self.regressor
            return x_projected
        else:
            return x
    
    def to_constraints(self):
        from ..functions import Variable, Abs
        x = Variable(self.input_shape[0])
        upper_bound = Abs(self.regressor @ x - self.shift) <= self.upper
        lower_bound = Abs(self.regressor @ x - self.shift) >= self.lower
        return [upper_bound, lower_bound]


class Circle(AbstractSet):
    """Circle set

    :math:`X = \\{x\\mid \\|x - c\\|<= r\\}`
    with :math:`x,c\\in \\mathbb{R}^{n}` and :math:`r\\in\\mathbb{R}`

    Args:
        center (numpy.ndarray): center of the circle
        radius (float): radius of the circle

    Attributes:
        center (numpy.ndarray): center of the circle
        radius (float): radius of the circle
        input_shape (tuple): space dimensions
    """

    def __init__(self, center, radius):
        if not isinstance(center, np.ndarray):
            raise ValueError("center must be numpy.ndarray")
        if len(center.shape) > 2 or center.shape[1] != 1:
            raise ValueError(
                "center must be 2D numpy.ndarray with shape (n,1), n>=1")
        self.center = center
        self.radius = radius
        self.input_shape = center.shape

    def __str__(self):
        description = r'\|x - {}\|<={}'.format(
            self.center.flatten(), self.radius)
        return description

    def intersection(self, circle):
        """Compute the intersection with another circle

        Args:
            circle (Circle): circle to compute the intersection with

        Raises:
            ValueError: Only intersection with another circle is supported
            ValueError: The two circles must have the same center
        """
        if not isinstance(circle, Circle):
            raise ValueError(
                "Only intersection with another circle is supported")
        if self.center.any() != circle.center.any():
            raise ValueError(
                "Circles must have the same center {}.".format(self.center))
        else:
            self.radius = np.min([self.radius, circle.radius])

    def projection(self, x):
        """Project a point onto the circle

        Arguments:
            x (numpy.ndarray): Point to be projected

        Returns:
            numpy.ndarray: projected point
        """
        if self.input_shape != x.shape:
            raise ValueError("shape of the point {} does not comply with the dimension of the circle {}".format(
                x.shape, self.input_shape))
        distance = np.linalg.norm(self.center - x)
        if distance > self.radius:
            x_projected = self.center + self.radius * \
                (x - self.center) / distance
            return x_projected
        else:
            return x

    def to_constraints(self):
        from ..functions import Variable, Norm
        x = Variable(self.input_shape[0])
        f = Norm(x - self.center) <= self.radius
        return [f]


class CircularSector(AbstractSet):
    """Circular sector set.

    Args:
        vertex (numpy.ndarray): vertex of the circular sector
        angle (float): direction of the circular sector (in rad)
        radius (float): radius of the circular sector
        width (float): width of the circular sector (in rad)

    Attributes:
        vertex (numpy.ndarray): vertex of the circular sector
        angle (float): direction of the circular sector (in rad)
        radius (float): radius of the circular sector
        h_angle (float): left border (from the vertex pov) of the circular sector (in rad)
        l_angle (float): right border (from the vertex pov) of the circular sector (in rad)
        input_shape (tuple): space dimensions
    """

    def __init__(self, vertex, angle, radius, width):
        if not isinstance(vertex, np.ndarray):
            raise ValueError("vertex must be a numpy.ndarray")
        if len(vertex.shape) > 2 or vertex.shape[1] != 1:
            raise ValueError(
                "vertex must be 2D numpy.ndarray with shape (n,1), n>=1")
        if vertex.shape[0] != 2:
            raise ValueError("only dimension (2,1) is curtrently supported")
        self.vertex = vertex
        self.radius = radius
        self.angle = angle
        self.h_angle = self.angle + width/2.0
        self.l_angle = self.angle - width/2.0
        self.input_shape = vertex.shape

    def __str__(self):
        return "Circular sector"  # TODO

    def intersection(self, circular_sector):
        """Compute the intersection with another circular sector

        Args:
            circular_sector (CircularSector): circula sector to compute the intersection with

        Raises:
            ValueError: Only intersection with another circular_sector is supported
            ValueError: The two circular_sector must have the same vertex
        """
        if not isinstance(circular_sector, CircularSector):
            raise ValueError(
                "Only intersection with another circular_sector is supported")
        if self.vertex.any() != circular_sector.vertex.any():
            raise ValueError("Circular sectors must have the same vertex.")
        else:
            self.radius = np.min([self.radius, circular_sector.radius])
            sl = np.unwrap(self.l_angle)
            cl = np.unwrap(circular_sector.l_angle)

            if self.h_angle >= 0:
                self.h_angle = np.min([self.h_angle, circular_sector.h_angle])
            elif self.h_angle < 0:
                if circular_sector.h_angle < 0:
                    self.h_angle = np.min(
                        [self.h_angle, circular_sector.h_angle])
                else:
                    self.h_angle = circular_sector.h_angle

            if self.l_angle >= 0:
                if circular_sector.l_angle >= 0:
                    self.l_angle = np.max(
                        [self.l_angle, circular_sector.l_angle])
                else:
                    if cl-sl < np.pi:
                        self.l_angle = circular_sector.l_angle
            if self.l_angle < 0:
                if circular_sector.l_angle < 0:
                    self.l_angle = np.max(
                        [self.l_angle, circular_sector.l_angle])
                else:
                    if cl-sl > np.pi:
                        self.l_angle = circular_sector.l_angle
            self.h_angle = np.array([self.h_angle]).flatten()
            self.l_angle = np.array([self.l_angle]).flatten()

    def __set_bounds(self):
        """set the bounds of the circular sector
        """
        self.a = np.array([-np.sin(self.h_angle), np.cos(self.h_angle)])
        self.a = self.a/np.linalg.norm(self.a)
        self.b = np.array([-np.sin(self.l_angle), np.cos(self.l_angle)])
        self.b = self.b/np.linalg.norm(self.b)
        self.ca = np.dot(self.a.T, self.vertex)
        self.cb = np.dot(self.b.T, self.vertex)

    def projection(self, x):
        """Project a point onto the ciruclar sector

        Arguments:
            x (numpy.ndarray): Point to be projected

        Returns:
            numpy.ndarray: projected point
        """
        if self.input_shape != x.shape:
            raise ValueError(
                "shape of the point {} does not comply with the dimension of the circular sector {}".format(
                    x.shape, self.input_shape))
        self.__set_bounds()
        v = x-self.vertex
        # if in cone remain the same
        if np.dot(self.a.T, v) <= 0 and np.dot(self.b.T, v) >= 0:
            x_projected = x
        # if behind the projection is the vertex
        elif np.dot(self.a.T, v) > 0 and np.dot(self.b.T, v) < 0:
            x_projected = self.vertex
        elif np.dot(self.a.T, v) > 0 and np.dot(self.b.T, v) >= 0:
            aa = self.a
            cc = np.dot(aa.T, self.vertex)
            x_projected = x+((cc-np.dot(aa.T, x)) /
                             np.linalg.norm(aa, ord=2)**2)*aa
            if np.dot(self.b.T, x_projected-self.vertex) < 0:
                x_projected = self.vertex
        elif np.dot(self.a.T, v) <= 0 and np.dot(self.b.T, v) < 0:
            bb = self.b
            cc = np.dot(bb.T, self.vertex)
            x_projected = x+((cc-np.dot(bb.T, x)) /
                             np.linalg.norm(bb, ord=2)**2)*bb
            if np.dot(self.a.T, x_projected-self.vertex) > 0:
                x_projected = self.vertex
        else:  # TODO check
            x_projected = x

        distance = np.linalg.norm(x_projected-self.vertex, ord=2)
        if distance > self.radius:
            x_projected = self.vertex+self.radius * \
                (x_projected-self.vertex)/distance
        return x_projected

    def to_constraints(self):
        from ..functions import Variable, Norm
        x = Variable(self.input_shape[0])
        circular_bound = Norm(x - self.vertex) <= self.radius
        upper_bound = self.a @ (x - self.vertex) <= 0
        lower_bound = self.b @ (x - self.vertex) >= 0
        return [circular_bound, upper_bound, lower_bound]




import numpy as np
import time
import warnings
import random
from typing import List, Tuple
from ..agents import Agent
from .algorithm import Algorithm


class Consensus(Algorithm):
    """Consensus Algorithm [OlSa07]_ 

    From the perspective of agent :math:`i` the algorithm works as follows. For :math:`k=0,1,\\dots`

    .. math::

        x_i^{k+1} = \sum_{j=1}^N w_{ij} x_j^k

    where :math:`x_i\\in\\mathbb{R}^n`. The weight matrix :math:`W=[w_{ij}]` should be doubly-stochastic in order to have convergence to the average of the local initial conditions. If :math:`W` is row-stochastic convergence is still attained but at a different point. Other type of matrices can be used, but convergence is not guaranteed.
    Also time-varying graphs can be adopted.

    Args:
        agent (Agent): agent to execute the algorithm
        initial_condition (numpy.ndarray): initial condition
        enable_log (bool): True for enabling log

    Attributes:
        agent (Agent): agent to execute the algorithm
        x0 (numpy.ndarray): initial condition
        x (numpy.ndarray): current value of the local solution
        shape (tuple): shape of the variable
        x_neigh (dict): dictionary containing the local solution of the (in-)neighbors
        enable_log (bool): True for enabling log
    """

    def __init__(self, agent: Agent, initial_condition: np.ndarray, enable_log: bool=False):
        super(Consensus, self).__init__(agent, enable_log)

        self.x0 = initial_condition
        self.x = initial_condition

        self.shape = self.x.shape

        self.x_neigh = {}

    def _update_local_solution(self, x: np.ndarray, **kwargs):
        """update the local solution

        Args:
            x: new value
 
        Raises:
            TypeError: Input must be a numpy.ndarray
            ValueError: Incompatible shapes
        """
        if not isinstance(x, np.ndarray):
            raise TypeError("Input must be a numpy.ndarray")
        if x.shape != self. x0.shape:
            raise ValueError("Incompatible shapes")
        self.x = x

    def iterate_run(self, **kwargs):
        """Run a single iterate of the algorithm
        """
        data = self.agent.neighbors_exchange(self.x)

        for neigh in data:
            self.x_neigh[neigh] = data[neigh]

        x_avg = self.agent.in_weights[self.agent.id] * self.x
        for i in self.agent.in_neighbors:
            x_avg += self.agent.in_weights[i] * self.x_neigh[i]

        self._update_local_solution(x_avg, **kwargs)

    def run(self, iterations: int=100, verbose: bool=False, **kwargs):
        """Run the algorithm for a given number of iterations

        Args:
            iterations: Number of iterations. Defaults to 100.
            verbose: If True print some information during the evolution of the algorithm. Defaults to False.
        """

        if not isinstance(iterations, int):
            raise TypeError("iterations must be an int")
        if self.enable_log:
            dims = [iterations]
            for dim in self.x.shape:
                dims.append(dim)
            self.sequence = np.zeros(dims)

        for k in range(iterations):
            self.iterate_run(**kwargs)

            if self.enable_log:
                self.sequence[k] = self.x
            
            if verbose:
                if self.agent.id == 0:
                    print('Iteration {}'.format(k), end="\r")

        if self.enable_log:
            return self.sequence

    def get_result(self):
        """Return the actual value of x

        Returns:
            numpy.ndarray: value of x
        """
        return self.x


class AsynchronousConsensus(Algorithm):
    """ Asynchronous Consensus Algorithm 

    From the perspective of agent :math:`i` the algorithm works as follows. When agent :math:`i` gets awake it updates its local solution as

    .. math::

        x_i \\gets \sum_{j\\in\\mathcal{N}_i} w_{ij} x_{j\\mid i}

    where :math:`\\mathcal{N}_i` is the current set of in-neighbors and :math:`x_{j\\mid i},j\\in\\mathcal{N}_i` is the local copy of :math:`x_j` available at node :math:`i` (which can be outdated, due to asynchrony, computation time and link failures).

    Args:
        agent: agent to execute the algorithm
        initial_condition: initial condition
        enable_log: True for enabling log. Defaults to False.
        force_sleep: True if one wanst to force sleep after the computation phase. Defaults to False.
        maximum_sleep: Maximum allowed sleep. Defaults to 0.01.
        sleep_type: Type of sleep time("constant", "random"). Defaults to "random".
        force_computation_time:  True if one want sto force length computation phase. Defaults to False.
        maximum_computation_time: Maximum allowed computation time. Defaults to 0.01.
        computation_time_type: Type of computation time ("constant", "random"). Defaults to "random".
        force_unreliable_links: True if one wants to force unreliable links. Defaults to False.
        link_failure_probability: Probability of incoming links failure. Defaults to 0.

    Attributes:
        agent (Agent): agent to execute the algorithm
        x0 (numpy.ndarray): initial condition
        x (numpy.ndarray): current value of the local solution
        shape (tuple): shape of the variable
        x_neigh (dict): dictionary containing the local solution of the (in-)neighbors
        enable_log (bool): True for enabling log
        timestamp_sequence_awake ( list ): list of timestamps at which node get awake
        timestamp_sequence_sleep ( list ): list of timestamps at which node go to sleep
        force_sleep: True if one wanst to force sleep after the computation phase. Defaults to False.
        maximum_sleep: Maximum allowed sleep. Defaults to 0.01.
        sleep_type: Type of sleep time("constant", "random"). Defaults to "random".
        force_computation_time:  True if one want sto force length computation phase. Defaults to False.
        maximum_computation_time: Maximum allowed computation time. Defaults to 0.01.
        computation_time_type: Type of computation time ("constant", "random"). Defaults to "random".
        force_unreliable_links: True if one wants to force unreliable links. Defaults to False.
        link_failure_probability: Probability of incoming links failure. Defaults to 0.
    """

    def __init__(self,
                 agent: Agent,
                 initial_condition: np.ndarray,
                 enable_log: bool = False,
                 force_sleep: bool = False,
                 maximum_sleep: float = 0.01,
                 sleep_type: str = "random",
                 force_computation_time: bool = False,
                 maximum_computation_time: float = 0.01,
                 computation_time_type: str = "random",
                 force_unreliable_links: bool = False,
                 link_failure_probability: float = 0):

        super(AsynchronousConsensus, self).__init__(agent, enable_log)

        self.x0 = initial_condition
        self.x = initial_condition

        self.shape = self.x.shape

        self.x_neigh = {}

        if force_sleep:
            if sleep_type not in ("constant", "random"):
                raise ValueError("sleep_type can be constant or random")
        if force_computation_time:
            if computation_time_type not in ("constant", "random"):
                raise ValueError("computation_time_type can be constant or random")
        self.enable_log = enable_log
        self.force_sleep = force_sleep
        self.maximum_sleep = maximum_sleep
        self.sleep_type = sleep_type
        self.force_computation_time = force_computation_time
        self.maximum_computation_time = maximum_computation_time
        self.computation_time_type = computation_time_type
        self.force_unreliable_links = force_unreliable_links
        self.link_failure_probability = link_failure_probability

        self.timestamp_sequence_awake = None
        self.timestamp_sequence_sleep = None

    def _update_local_solution(self, x: np.ndarray, **kwargs):
        """update the local solution

        Args:
            x: new value

        Raises:
            TypeError: Input must be a numpy.ndarray
            ValueError: Incompatible shapes
        """
        if not isinstance(x, np.ndarray):
            raise TypeError("Input must be a numpy.ndarray")
        if x.shape != self. x0.shape:
            raise ValueError("Incompatible shapes")
        self.x = x

    def iterate_run(self, **kwargs):
        """Run a single iterate
        """

        data = self.agent.neighbors_receive_asynchronous()

        if self.enable_log:
            self.timestamp_sequence_awake.append(time.time())

        for neigh in data:
            self.x_neigh[neigh] = data[neigh]

        if self.force_computation_time:
            start_wait_time = time.time()

        if not self.force_unreliable_links:
            x_avg = self.agent.in_weights[self.agent.id] * self.x
            for i in self.agent.in_neighbors:
                x_avg += self.agent.in_weights[i] * self.x_neigh[i]
        else:
            warnings.warn("when forcing link failures, all neighbors are given the same weight")
            neighbors = []
            for neigh in self.agent.in_neighbors:
                rnd = random.uniform(0, 1)
                if rnd > self.link_failure_probability:
                    neighbors.append(neigh)
            weight = 1.0/(len(neighbors) + 1)
            x_avg = weight * self.x
            for i in neighbors:
                x_avg += weight * self.x_neigh[i]

        self._update_local_solution(x_avg, **kwargs)

        # save sequence
        if self.enable_log:
            self.sequence = np.vstack([self.sequence, self.x.reshape(self.dims)])

        # force computation time if requested
        if self.force_computation_time:
            if self.computation_time_type == "random":
                wait_time = random.uniform(0, self.maximum_computation_time)
            elif self.computation_time_type == "constant":
                wait_time = self.maximum_computation_time
            remaining_wait_time = wait_time - (time.time() - start_wait_time)
            if remaining_wait_time < 0:
                warnings.warn("requested computation time cannot be guaranteed")
            else:
                time.sleep(remaining_wait_time)

        self.agent.neighbors_send(self.x)
        if self.enable_log:
            self.timestamp_sequence_sleep.append(time.time())

    def run(self, running_time: float = 5.0):
        """Run the asynchronous consensus algorithm for a certain amount of time

        Args:
            running_time: Total run time. Defaults to 5.0.

        Returns:
            tuple: timestamp_sequence_awake, timestamp_sequence_sleep, sequence
        """
        if not isinstance(running_time, (int, float)):
            raise TypeError("Running time must be a float")
        if self.enable_log:
            dims = [1]
            for dim in self.x.shape:
                dims.append(dim)
            self.dims = dims
            self.sequence = np.zeros(dims)
            self.sequence[0] = self.x
            self.timestamp_sequence_awake = [time.time()]
            self.timestamp_sequence_sleep = [time.time()]

        # Exchange all data at the beginning
        data = self.agent.neighbors_exchange(self.x)
        for neigh in data:
            self.x_neigh[neigh] = data[neigh]

        # Then go asyncrhronous
        start_time = time.time()
        end_time = start_time + running_time
        while time.time() < end_time:
            self.iterate_run()

            # force a sleep time if requested
            if self.force_sleep:
                start_sleep_time = time.time()
                if self.sleep_type == "random":
                    sleep_time = random.uniform(0, self.maximum_sleep)
                elif self.sleep_type == "constant":
                    sleep_time = self.maximum_sleep

                remaining_sleep_time = sleep_time - \
                    (time.time() - start_sleep_time)
                if remaining_sleep_time < 0:
                    warnings.warn("requested delay cannot be guaranteed")
                else:
                    time.sleep(remaining_sleep_time)

        if self.enable_log:
            return self.timestamp_sequence_awake, \
                self.timestamp_sequence_sleep, \
                self.sequence

    def get_result(self):
        """Return the actual value of x

        Returns:
            numpy.ndarray: value of x
        """
        return self.x


class BlockConsensus(Algorithm):
    """Block-wise consensus [FaNo19]_

    At each iteration, the agent can update its local estimate or not at each iteration according to a certain probability (awakening_probability).
    From the perspective of agent :math:`i` the algorithm works as follows. At iteration :math:`k` if the agent is awake, it selects a random block :math:`\\ell_i^k` of its local solution and updates

    .. math::

        x_{i,\\ell}^{k+1} = \\begin{cases}
             \\sum_{j\\in\\mathcal{N}_i} w_{ij} x_{j\\mid i,\\ell}^k & \\text{if} \\ell = \\ell_i^k \\\\
             x_{i,\\ell}^{k} & \\text{otherwise}
             \\end{cases}

    where :math:`\\mathcal{N}_i` is the current set of in-neighbors and :math:`x_{j\\mid i},j\\in\\mathcal{N}_i` is the local copy of :math:`x_j` available at node :math:`i` and :math:`x_{i,\\ell}` denotes the :math:`\\ell`-th block of :math:`x_i`. Otherwise :math:`x_{i}^{k+1}=x_i^k`.



    Args:
        agent: agent to execute the algorithm
        initial_condition: initial condition
        enable_log: True for enabling log
        blocks_list: the list of blocks (list of tuples)
        probabilities: list of probabilities of drawing each block
        awakening_probability: probability of getting awake at each iteration

    """

    def __init__(self,
                 agent: Agent,
                 initial_condition: np.ndarray,
                 enable_log: bool = False,
                 blocks_list: List[Tuple] = None,
                 probabilities: List[float] = None,
                 awakening_probability: float = 1.0):

        super(BlockConsensus, self).__init__(agent, enable_log)
        if (not isinstance(initial_condition, np.ndarray)) or \
            (len(initial_condition.shape) != 2) or \
                (initial_condition.shape[1] != 1):
            raise ValueError("Initial condition must be a numpy.ndarray with shape (Any, 1)")

        self.x0 = initial_condition
        self.x = initial_condition

        self.shape = self.x.shape

        self.x_neigh = {}
        if blocks_list is not None:
            if not isinstance(blocks_list, list):
                raise ValueError("blocks_list argument, if provided, must be a list of tuples")
            items = 0
            for item in blocks_list:
                if not isinstance(item, tuple):
                    raise ValueError("blocks_list argument, if provided, must be a list of tuples")
                items += len(item)
            self.blocks_list = blocks_list
            if items != self.shape[0]:
                warnings.warn("Not all elements have been included in blocks_list")
        else:
            self.blocks_list = list(range(self.shape[0]))

        self.blocks_number = len(self.blocks_list)

        if probabilities is not None:
            if isinstance(probabilities, list):
                for i in probabilities:
                    if not isinstance(i, float):
                        raise ValueError("probabilities argument, if provided, must be a list of float")
            else:
                raise ValueError("probabilities argument, if provided, must be a list of float")

            if len(probabilities) != len(self.blocks_list):
                raise ValueError("blocks_list and probabilities arguments have different lengths")

            if np.sum(probabilities) != 1.0:
                raise ValueError("probabilities must sum to 1")

            self.probabilities = probabilities
        else:
            self.probabilities = (np.ones(self.shape).flatten()/self.shape[0]).tolist()

        if isinstance(awakening_probability, float) and (0 <= awakening_probability <= 1):
            self.awakening_probability = awakening_probability
        else:
            raise ValueError("awakening_probability must be a float in [0,1]")

    def _update_local_solution(self, x: np.ndarray, **kwargs):
        """update the local solution
        
        Args:
            x: new value
        
        Raises:
            TypeError: Input must be a numpy.ndarray
            ValueError: Incompatible shapes
        """
        if not isinstance(x, np.ndarray):
            raise TypeError("Input must be a numpy.ndarray")
        if x.shape != self. x0.shape:
            raise ValueError("Incompatible shapes")
        self.x = x

    def iterate_run(self, **kwargs):
        """Run a single iterate of the algorithm
        """
        awake = random.uniform(0, 1)
        if awake <= self.awakening_probability:
            selected_index = np.random.choice(np.arange(self.blocks_number), p=self.probabilities)
            
            selected_block = self.blocks_list[selected_index]
            if isinstance(selected_block, int):
                selected_block = (selected_block, )
            packet_send = {'block': selected_block, 'data': self.x[list(selected_block)]}
            data = self.agent.neighbors_exchange(packet_send)

            for neigh in data:
                received_block = data[neigh]['block']
                received_data = data[neigh]['data']
                try:
                    self.x_neigh[neigh][list(received_block)] = received_data
                except KeyError:
                    self.x_neigh[neigh] = np.zeros(self.shape)
                    self.x_neigh[neigh][list(received_block)] = received_data

            x_avg = self.agent.in_weights[self.agent.id] * self.x
            for i in self.agent.in_neighbors:
                x_avg += self.agent.in_weights[i] * self.x_neigh[i]

            self._update_local_solution(x_avg, selected_block=selected_block, **kwargs)

    def run(self, iterations: int=100, verbose: bool=False):
        """Run the algorithm for a given number of iterations

        Args:
            iterations: Number of iterations. Defaults to 100.
            verbose: If True print some information during the evolution of the algorithm. Defaults to False.
        """
        if not isinstance(iterations, int):
            raise TypeError("iterations must be an int")
        if self.enable_log:
            dims = [iterations]
            for dim in self.x.shape:
                dims.append(dim)
            self.sequence = np.zeros(dims)

        for k in range(iterations):
            self.iterate_run()

            if self.enable_log:
                self.sequence[k] = self.x
            
            if verbose:
                if self.agent.id == 0:
                    print('Iteration {}'.format(k), end="\r")

        if self.enable_log:
            return self.sequence

    def get_result(self):
        """Return the actual value of x

        Returns:
            numpy.ndarray: value of x
        """
        return self.x


class PushSumConsensus(Algorithm):
    """Push-Sum Consensus Algorithm 

    From the perspective of agent :math:`i` the algorithm works as follows. For :math:`k=0,1,\\dots`

    .. math::

        x_i^{k+1} &= \\sum_{j=1}^N w_{ij} x_j^k

        y_i^{k+1} &= \\sum_{j=1}^N w_{ij} y_j^k

        z_i^{k+1} &= \\frac{x_i^{k+1}}{y_i^{k+1}}

    where :math:`x_i\\in\\mathbb{R}^n`. The weight matrix :math:`W=[w_{ij}]` should be column-stochastic in order to let :math:`z_i^k` converge to the average of the local initial conditions.
    Also time-varying graphs can be adopted.


    Args:
        agent (Agent): agent to execute the algorithm
        initial_condition (numpy.ndarray): initial condition
        enable_log (bool): True for enabling log

    Attributes:
        agent (Agent): agent to execute the algorithm
        z0 (numpy.ndarray): initial condition
        z (numpy.ndarray): current value of the local solution
        shape (tuple): shape of the variable
        x_neigh (dict): dictionary containing the x values of the (in-)neighbors
        y_neigh (dict): dictionary containing the y values of the (in-)neighbors
        enable_log (bool): True for enabling log
    """

    def __init__(self, agent: Agent, initial_condition: np.ndarray, enable_log: bool=False):
        super(PushSumConsensus, self).__init__(agent, enable_log)

        self.z0 = initial_condition
        self.x = initial_condition
        self.z = initial_condition
        self.y = np.ones(initial_condition.shape)

        self.shape = self.z0.shape

        self.x_neigh = {}
        self.y_neigh = {}

    def _update_x_average(self, x: np.ndarray, **kwargs):
        if not isinstance(x, np.ndarray):
            raise TypeError("Input must be a numpy.ndarray")
        if x.shape != self.x.shape:
            raise ValueError("Incompatible shapes")
        self.x = x
    
    def _update_y_average(self, y: np.ndarray, **kwargs):
        if not isinstance(y, np.ndarray):
            raise TypeError("Input must be a numpy.ndarray")
        if y.shape != self.y.shape:
            raise ValueError("Incompatible shapes")
        self.y = y

    def _update_local_solution(self, z: np.ndarray, **kwargs):
        """update the local solution
        
        Args:
            x: new value
        
        Raises:
            TypeError: Input must be a numpy.ndarray
            ValueError: Incompatible shapes
        """
        if not isinstance(z, np.ndarray):
            raise TypeError("Input must be a numpy.ndarray")
        if z.shape != self.z.shape:
            raise ValueError("Incompatible shapes")
        self.z = z

    def iterate_run(self, **kwargs):
        """Run a single iterate of the algorithm
        """
        # # in
        # data = self.agent.neighbors_exchange(self.x)

        # for neigh in data:
        #     self.x_neigh[neigh] = data[neigh]

        # x_avg = self.agent.in_weights[self.agent.id] * self.x
        # for i in self.agent.in_neighbors:
        #     x_avg += self.agent.in_weights[i] * self.x_neigh[i]

        # x average
        send_data = {}
        for j in self.agent.out_neighbors:
            send_data[j] = self.agent.out_weights[j] * self.x
        data = self.agent.neighbors_exchange(send_data, dict_neigh=True)
        
        for neigh in data:
            self.x_neigh[neigh] = data[neigh]

        x_avg = self.agent.out_weights[self.agent.id] * self.x
        for i in self.agent.in_neighbors:
            x_avg += self.x_neigh[i]

        self._update_x_average(x_avg, **kwargs)

        # y average
        send_data = {}
        for j in self.agent.out_neighbors:
            send_data[j] = self.agent.out_weights[j] * self.y
        data = self.agent.neighbors_exchange(send_data, dict_neigh=True)

        for neigh in data:
            self.y_neigh[neigh] = data[neigh]

        y_avg = self.agent.out_weights[self.agent.id] * self.y
        for i in self.agent.in_neighbors:
            y_avg += self.y_neigh[i]

        self._update_y_average(y_avg, **kwargs)

        # aggregate
        z = self.x/self.y
        self._update_local_solution(z)

    def run(self, iterations: int=100, verbose: bool=False, **kwargs):
        """Run the algorithm for a given number of iterations

        Args:
            iterations: Number of iterations. Defaults to 100.
            verbose: If True print some information during the evolution of the algorithm. Defaults to False.
        """
        if not isinstance(iterations, int):
            raise TypeError("iterations must be an int")
        if self.enable_log:
            dims = [iterations]
            for dim in self.x.shape:
                dims.append(dim)
            self.sequence = np.zeros(dims)

        for k in range(iterations):
            self.iterate_run(**kwargs)

            if self.enable_log:
                self.sequence[k] = self.x
            
            if verbose:
                if self.agent.id == 0:
                    print('Iteration {}'.format(k), end="\r")

        if self.enable_log:
            return self.sequence

    def get_result(self):
        """Return the actual value of x

        Returns:
            numpy.ndarray: value of x
        """
        return self.z



from .abstract_function import AbstractFunction
from .extended_function import ExtendedFunction
from .variable import Variable
from .affine_form import AffineForm
from .quadratic_form import QuadraticForm
from .norm import Norm
from .squared_norm import SquaredNorm
from .exp import Exp
from .log import Log
from .logistic import Logistic
from .max import Max
from .min import Min
from .power import Power
from .square import Square
from .abs import Abs



import numpy as np
import warnings
from .abstract_function import AbstractFunction
from .utilities import check_input
from .max import Max


class Min(Max):
    """Min function (elementwise)

    .. math::

        f(x,y) = \\min(x,y)

    with :math:`x,y: \\mathbb{R}^{n}`.

    Args:
        f1 (AbstractFunction): input function
        f2 (AbstractFunction): input function

    Raises:
        ValueError: input must be a AbstractFunction object
        ValueError: sunctions must have the same input/output shapes
    """

    def __init__(self, f1: AbstractFunction, f2: AbstractFunction):
        super(Min, self).__init__(-f1, -f2)

    def _to_cvxpy(self):
        import cvxpy as cvx
        return cvx.minimum(self.f1._to_cvxpy(), self.f2._to_cvxpy())
    
    def _extend_variable(self, n_var, axis, pos):
        return Min(-self.f1._extend_variable(n_var, axis, pos), -self.f2._extend_variable(n_var, axis, pos))

    @check_input
    def eval(self, x: np.ndarray) -> np.ndarray:
        return -super(Min, self).eval(x)

    # @check_input
    # def jacobian(self, x: np.ndarray, **kwargs) -> np.ndarray:
    #     return -super(Min, self).jacobian(x, **kwargs)




import numpy as np
from ..functions.utilities import check_input


class AbstractConstraint:
    """Abstract class for expressing constraints
    """
    def eval(self):
        pass


class Constraint(AbstractConstraint):
    """Constraint build from a AbstractFunction object. Constraints are represented in the canonical forms :math:`f(x)=0` and :math:`f(x)\leq 0`.

    Args:
        fn (AbstractFunction): constraint function 
        sign (bool): type of constraint: "==", "<=" or ">="

    Attributes:
        fn (AbstractFunction): constraint function 
        sign (bool): type of constraint: "==", "<=" or ">="
        input_shape (tuple): input space dimensions
        output_shape (tuple): output space dimensions
    """

    def __init__(self, fn, sign: str = "=="):
        self.input_shape = fn.input_shape
        self.output_shape = fn.output_shape
        self.sign = sign
        self.fn = fn

        if self.sign == ">=":
            self.sign = "<="
            self.fn = -fn
    
    def _to_cvxpy(self):
        if self.is_equality:
            return self.fn._to_cvxpy() == 0
        else:
            return self.fn._to_cvxpy() <= 0
            
    @property
    def is_equality(self):
        return self.sign == "=="
    
    @property
    def is_inequality(self):
        return self.sign == "<="

    @property
    def is_affine(self):
        """Return true if the function is affine.

        Returns:
            bool: true if the function is affine
        """
        return self.fn.is_affine

    @property
    def is_quadratic(self):
        """Return true if the function is affine.

        Returns:
            bool: true if the function is affine
        """
        return self.fn.is_quadratic

    @property
    def function(self):
        return self.fn

    def get_parameters(self):
        """Return the parameters of the function if it is affine or quadratic

        Returns:
            tuple: A, b for affine constraints, P, q, r for quadratic
        """
        return self.fn.get_parameters()

    @check_input
    def eval(self, x: np.ndarray) -> bool:
        """Evaluate the constraint function at a point x

        Args:
            x: input point
        """
        if self.is_equality:
            return np.allclose(self.fn.eval(x), np.zeros(self.output_shape))
        elif self.is_inequality:
            if np.allclose(self.fn.eval(x), np.zeros(self.output_shape)):
                return True
            return np.less_equal(self.fn.eval(x), np.zeros(self.output_shape)).all()

    @check_input
    def projection(self, x: np.ndarray) -> np.ndarray:
        """Compute the projection of a point onto the set defined by the constraint. The constraint should be convex.
        
        Args:
            x : point to be projected
        
        Returns:
            numpy.ndarray: projected point
        """
        satisfied = self.eval(x)
        if satisfied:
            return x
        else:
            if self.is_affine:
                A, b = self.get_parameters()
                if A.shape[0] >= A.shape[1]:
                    projected_pt = x - A @ np.linalg.inv(A.transpose() @ A) @ (A.transpose() @ x + b)
                    return projected_pt
                else:
                    from ..problems import Problem
                    from ..functions import Variable, QuadraticForm
                    v = Variable(self.input_shape[0])
                    P = np.eye(self.input_shape[0])
                    obj = 0.5 * QuadraticForm(v - x, P)
                    constraint = self
                    pb = Problem(obj, constraint)
                    projected_pt = pb.solve()
                    return projected_pt
            else:
                from ..problems import Problem
                from ..functions import Variable, QuadraticForm
                v = Variable(self.input_shape[0])
                P = np.eye(self.input_shape[0])
                obj = 0.5 * QuadraticForm(v - x, P)
                constraint = self
                pb = Problem(obj, constraint)
                projected_pt = pb.solve()
                return projected_pt




import numpy as np
import warnings
from typing import Union
from ..functions import AbstractFunction
from ..functions.affine_form import aggregate_affine_form
from ..constraints import Constraint, AbstractSet
from .utilities import check_affine_constraints
from ..utils.utilities import check_symmetric, is_semi_pos_def


class Problem:
    """A generic optimization problem. 

    .. math::

        \\text{minimize } & f(x)

        \\text{subject to } & g(x) \\leq 0

                            & h(x) = 0

    Args:
        objective_function (AbstractFunction, optional): objective function. Defaults to None.
        constraints (list, optional): constraints. Defaults to None.

    Attributes:
        objective_function (Function): Objective function to be minimized
        constraints (list): constraints
        input_shape (tuple): dimension of optimization variable
        output_shape (tuple): output shape
    """

    def __new__(cls, objective_function: AbstractFunction = None, constraints: list = None, force_general_problem: bool = False, **kwargs):
        instance = object.__new__(cls)
        if not force_general_problem:
            if cls.__name__ == 'Problem':
                if objective_function is not None:  # TODO feasibility problem
                    if objective_function.is_affine:
                        if check_affine_constraints(constraints):
                            # Convert to LinearProblem
                            from .linear_problem import LinearProblem
                            instance = LinearProblem(objective_function, constraints)

                    elif objective_function.is_quadratic:
                        P, _, _ = objective_function.get_parameters()
                        if is_semi_pos_def(P):
                            if check_affine_constraints(constraints):
                                # Convert to QuadraticProblem
                                from .quadratic_problem import QuadraticProblem
                                instance = QuadraticProblem(objective_function, constraints, is_pos_def=True)
        return instance

    def __init__(self, objective_function: AbstractFunction = None, constraints: list = None,
                 force_general_problem: bool = False, **kwargs):
        self.objective_function = None
        self.constraints = []
        self.input_shape = None
        self.output_shape = None
        if objective_function is not None:
            self.set_objective_function(objective_function)

        if constraints is not None:
            if not isinstance(constraints, list):
                constraints = [constraints]
            for constraint in constraints:
                self.add_constraint(constraint)

        self.__split_constraints()

    def set_objective_function(self, fn: AbstractFunction):
        """Set the objective function

        Args:
            fn: objective function

        Raises:
            TypeError: input must be a AbstractFunction object
        """
        if isinstance(fn, AbstractFunction):
            self.objective_function = fn
            self.input_shape = fn.input_shape
            self.output_shape = fn.output_shape
        else:
            raise TypeError("input must be a AbstractFunction")

    def add_constraint(self, fn: Union[AbstractSet, Constraint]):
        """Add a new constraint

        Args:
            fn: new constraint

        Raises:
            TypeError: constraints must be AbstractSet or Constraint
        """
        if isinstance(fn, Constraint):
            self.constraints.append(fn)
        elif isinstance(fn, AbstractSet):
            constraints_list = fn.to_constraints()
            for const in constraints_list:
                self.constraints.append(const)
        else:
            raise TypeError("constraints must be AbstractSet or Constraint")

    def __split_constraints(self):
        """split the constraints into affine equalities, affine inequalities and nonlinear constraints

        Raises:
            ValueError: Nonlinear constraints must have scalar output
        """
        affine_equalities = []
        affine_inequalities = []
        self.affine_equality = None
        self.affine_inequality = None
        self.nonlinear_constraints = []
        self.constraints_map = {}
        equality_idx_counter = 0
        inequality_idx_counter = 0
        nonlinear_idx_counter = 0
        for idx, constr in enumerate(self.constraints):
            if constr.is_affine:
                if constr.is_equality:
                    affine_equalities.append(constr.function)
                    self.constraints_map[idx] = {'type': 'affine_equality', 'indices': list(
                        range(equality_idx_counter, equality_idx_counter+constr.output_shape[0]))}
                    equality_idx_counter += constr.output_shape[0]
                else:
                    affine_inequalities.append(constr.function)
                    self.constraints_map[idx] = {'type': 'affine_inequality', 'indices': list(
                        range(inequality_idx_counter, inequality_idx_counter+constr.output_shape[0]))}
                    inequality_idx_counter += constr.output_shape[0]
            else:
                if constr.output_shape != (1, 1):
                    raise ValueError("Nonlinear constraints must have scalar output")
                if constr.is_inequality:
                    self.nonlinear_constraints.append(constr)
                    self.constraints_map[idx] = {'type': 'nonlinear_inequality', 'indices': list(
                        range(nonlinear_idx_counter, nonlinear_idx_counter+constr.output_shape[0]))}
                    nonlinear_idx_counter += constr.output_shape[0]
                else:
                    warnings.warn("Nonconvex problem")
                    c1 = constr.function <= 0
                    c2 = constr.function >= 0
                    self.nonlinear_constraints.append(c1)
                    self.nonlinear_constraints.append(c2)

        if len(affine_equalities) != 0:
            aggregated_equality = aggregate_affine_form(affine_equalities)
            self.affine_equality = aggregated_equality == 0

        if len(affine_inequalities) != 0:
            aggregated_inequality = aggregate_affine_form(affine_inequalities)
            self.affine_inequality = aggregated_inequality <= 0

    def __nonlinear_f_cvxopt(self, x=None, z=None):
        """nonlinear F function for cvxopt solver
        """
        try:
            import cvxopt
        except ImportError:
            ImportError("CVXOPT is required")

        m = len(self.nonlinear_constraints)
        if x is None and z is None:
            # TODO Check domain of functions
            x0 = np.random.randn(*self.input_shape)
            return m, cvxopt.matrix(x0)

        if z is None:
            # TODO check size of x=(n, 1)
            x_np = np.array(x)
            f = np.zeros([m + 1, 1])
            Df = np.zeros([m + 1, self.input_shape[0]])
            f[0] = self.objective_function.eval(x_np)
            Df[0, :] = self.objective_function.subgradient(x_np).flatten()
            for idx, constr in enumerate(self.nonlinear_constraints):
                f[idx + 1] = constr.function.eval(x_np)
                Df[idx + 1, :] = constr.function.subgradient(x_np).flatten()
            # TODO check domain and return None if x is not in the domain
            return cvxopt.matrix(f), cvxopt.matrix(Df)

        # TODO check size of x=(n, 1) and z=(m+1, 1)
        x_np = np.array(x)
        f = np.zeros([m + 1, 1])
        Df = np.zeros([m + 1, self.input_shape[0]])
        H = np.zeros([self.input_shape[0], self.input_shape[0]])
        f[0] = self.objective_function.eval(x_np)
        Df[0, :] = self.objective_function.subgradient(x_np).flatten()
        H += z[0] * self.objective_function.hessian(x_np)
        for idx, constr in enumerate(self.nonlinear_constraints):
            f[idx + 1] = constr.function.eval(x_np)
            Df[idx + 1, :] = constr.function.subgradient(x_np).flatten()
            H += z[idx+1] * constr.function.hessian(x_np)
        return cvxopt.matrix(f), cvxopt.matrix(Df), cvxopt.matrix(H)

    def solve(self, solver='cvxpy', return_only_solution: bool = True):
        """Solve the problem

        Returns:
            numpy.ndarray: solution
        """
        if solver == 'cvxpy':
            try:
                import cvxpy as cvx
            except ImportError:
                warnings.warn("CVXPY is not installed. Trying with CVXOPT.")
                solver = 'cvxopt'
            try:
                cvxpy_func = self.objective_function._to_cvxpy()
                obj = cvx.Minimize(cvxpy_func)
                constraints = []
                for const in self.constraints:
                    constraints.append(const._to_cvxpy())
                pb = cvx.Problem(obj, constraints)
                try:
                    pb.solve()
                except cvx.SolverError:
                    pb.solve(solver='SCS')
                except RuntimeError:
                    pb.solve(solver='ECOS')

                sol = np.array(pb.variables()[0].value).reshape(self.input_shape)
                if not return_only_solution:
                    if pb.status == 'optimal':
                        status = 'solved'
                    else:
                        status = pb.status
                    dual_variables = {}
                    for idx, const in enumerate(constraints):
                        dual_variables[idx] = np.array(const.dual_value).reshape(self.constraints[idx].output_shape)

                    output_dict = {
                        'solution': sol,
                        'status': status,
                        'dual_variables': dual_variables
                    }
                    return output_dict
                else:
                    return sol
            except RuntimeError:
                warnings.warn("CVXPY cannot solve the problem. Trying with CVXOPT.")
                solver = 'cvxopt'
            except:
                warnings.warn("CVXPY cannot solve the problem. Trying with CVXOPT.")
                solver = 'cvxopt'

        if solver == 'cvxopt':
            try:
                import cvxopt
            except ImportError:
                raise ImportError("CVXOPT is required")
            cvxopt.solvers.options['show_progress'] = False
            cvxopt.solvers.options['maxiters'] = 1000
            try:
                F = self.__nonlinear_f_cvxopt
                G = None
                h = None
                A = None
                b = None
                if self.affine_inequality is not None:
                    G, h = self.affine_inequality.get_parameters()
                    G = G.transpose()
                    G = cvxopt.matrix(G)
                    h = cvxopt.matrix(-h)
                if self.affine_equality is not None:
                    A, b = self.affine_equality.get_parameters()
                    A = A.transpose()
                    A = cvxopt.matrix(A)
                    b = cvxopt.matrix(-b)
                sol = cvxopt.solvers.cp(F, G=G, h=h, A=A, b=b)

                if not return_only_solution:
                    dual_variables = {}
                    for idx, _ in enumerate(self.constraints):
                        if self.constraints_map[idx]['type'] == 'affine_equality':
                            dual_variables[idx] = np.array(sol['y'])[self.constraints_map[idx]['indices']]
                        elif self.constraints_map[idx]['type'] == 'affine_inequality':
                            dual_variables[idx] = np.array(sol['zl'])[self.constraints_map[idx]['indices']]
                        elif self.constraints_map[idx]['type'] == 'nonlinear_inequality':
                            dual_variables[idx] = np.array(sol['znl'])[self.constraints_map[idx]['indices']]

                    if sol['status'] == 'optimal':
                        status = 'solved'
                    else:
                        status = sol['status']

                    output_dict = {
                        'solution': np.array(sol['x']).reshape(self.input_shape),
                        'status': status,
                        'dual_variables': dual_variables
                    }
                    return output_dict
                else:
                    return np.array(sol['x']).reshape(self.input_shape)
            except:
                raise ValueError("Cannot solve the problem")

    def project_on_constraint_set(self, x: np.ndarray) -> np.ndarray:
        """Compute the projection of a point onto the constraint set of the problem

        Args:
            x : point to project

        Returns:
            numpy.ndarray: projected point
        """
        from .projection_problem import ProjectionProblem
        pb = ProjectionProblem(self.constraints, x)
        return pb.solve()



import numpy  as np
import warnings
from .abstract_function import AbstractFunction
from .utilities import check_input

# DEPRECATED

class Constant(AbstractFunction):
    """Constant, basic function

    .. math::

        f(x)=c

    with :math:`c\\in \\mathbb{R}^{n}`.

    Args:
        c (numpy.ndarray): constant value (with shape (n, 1))
    """

    differentiable = True
    affine = True

    def __init__(self, c: np.ndarray):
        if not isinstance(c, np.ndarray):
            raise TypeError("Input must be an numpy.ndarray")
        if len(c.shape) != 2 or c.shape[1] != 1:
            raise ValueError("Input must be an numpy.ndarray with shape (Any, 1)")
        self.value = c
        self.input_shape = c.shape
        self.output_shape = c.shape

    #@check_input
    def eval(self, x: np.ndarray=None) -> np.ndarray:
        return self.value.reshape(self.output_shape)

    #@check_input
    def jacobian(self, x: np.ndarray, **kwargs) -> np.ndarray:
        return np.zeros([self.input_shape[0], x.shape[0]])



import numpy as np
from copy import deepcopy
from scipy import stats
import time
import warnings
import random
from typing import Optional, List, Tuple, Any
from ..agents import Agent
from .algorithm import Algorithm
from .misc import AsynchronousLogicAnd, LogicAnd
from ..utils.utilities import is_pos_def
from ..functions import Variable, SquaredNorm, Square, Max, Norm, Square, Abs

class ASYMM(AsynchronousLogicAnd):
    """Asyncrhonous Distributed Method of Multipliers [FaGa19b]_

    See [FaGa19b]_ for the details.

    .. warning:: 
        This algorithm is currently under development

    """

    def __init__(self,
                 agent: Agent,
                 graph_diameter: int,
                 initial_condition: np.ndarray,
                 enable_log: bool = False,
                 **kwargs):
        super(ASYMM, self).__init__(agent=agent,
                                    graph_diameter=graph_diameter,
                                    enable_log=enable_log,
                                    **kwargs)

        self.x0 = initial_condition
        self.x_old = initial_condition
        self.x = initial_condition

        self.shape = self.x.shape

        self.x_neigh = {}

        # multipliers
        self.nu_from = {}
        self.nu_to = {}
        self.rho_from = {}
        self.rho_to = {}
        self.gr_cn = {}
        self.multiplier = {}
        self.penalty = {}
        self.constr_val = {}

        # Algorithm parameters
        # max penalty
        self.max_penalty = 1e5
        # constant stepsize
        self.alpha = 0.1
        # initial tolerance
        self.e = 0.1
        # penalty threshold
        self.gamma = 0.25
        # penalty growth parameter
        self.beta = 2

        # auxiliary variables
        self.M_done = False

        # local lagrangian
        self.local_lagrangian = None

    def __initialize_multipliers(self):
        for neigh in self.agent.in_neighbors:
            self.nu_from[neigh] = 0.000001*np.random.rand(*self.shape)
            self.rho_from[neigh] = 0.1
            self.nu_to[neigh] = 0.000001*np.random.rand(*self.shape)
            self.rho_to[neigh] = 0.1

            self.gr_cn[neigh] = np.linalg.norm(self.x - self.x_neigh[neigh], ord=2)
        for idx, constraint in enumerate(self.agent.problem.constraints):
            self.multiplier[idx] = 0.000001*np.random.rand(*constraint.output_shape)
            self.penalty[idx] = 0.1

            if constraint.is_equality:
                self.constr_val[idx] = np.linalg.norm(constraint.function.eval(self.x), ord=2)
            elif constraint.is_inequality:
                self.constr_val[idx] = np.linalg.norm(
                    np.max([-self.multiplier[idx]/self.penalty[idx], constraint.function.eval(self.x)], axis=0))

    def __update_local_lagrangian(self):
        x = Variable(self.shape[0])

        # objective function
        fn = 0
        fn += self.agent.problem.objective_function

        # neighboring constraints
        for j in self.agent.in_neighbors:
            fn += (self.nu_to[j] - self.nu_from[j]) @ x + (self.rho_to[j]+self.rho_from[j]) / 2 * (x - self.x_neigh[j]) @ (x - self.x_neigh[j])
            # fn += (self.rho_to[j]+self.rho_from[j]) / 2 * Abs((x - self.x_neigh[j])@(x - self.x_neigh[j]) - 1e-7)

        # equality constraints
        for idx, constraint in enumerate(self.agent.problem.constraints):
            if constraint.is_equality:
                fn += self.multiplier[idx] @ constraint.function + \
                    0.5 * self.penalty[idx] * SquaredNorm(constraint.function)
            elif constraint.is_inequality:
                shape = constraint.output_shape
                fn += 1/(2*self.penalty[idx]) * np.ones(shape) @ (
                    Square(Max(0, self.multiplier[idx] + self.penalty[idx]*constraint.function)) - np.power(self.multiplier[idx], 2))

        self.local_lagrangian = fn

    def __update_stepsize(self):
        # TODO stepsize estimate
        self.alpha /= 1.4

    def __perform_descent(self, stepsize_type='newton'):
        if stepsize_type == 'newton':
            gradient = self.local_lagrangian.subgradient(self.x)
            hessian = self.local_lagrangian.hessian(self.x)
            if is_pos_def(hessian):
                try:
                    self.x -= 0.5 * np.linalg.inv(hessian) @ gradient
                except:
                    self.x -= 0.5 * 1/np.linalg.norm(hessian) * gradient
            else:
                self.x -= min(self.alpha, self.e/10) * gradient
        else:
            self.x -= self.alpha * self.local_lagrangian.subgradient(self.x)
    
    def primal_update_step(self):
        self.__update_local_lagrangian()
        self.__perform_descent()
        gradient = self.local_lagrangian.subgradient(self.x)
        if np.linalg.norm(gradient, ord=2) <= self.e:
            self.change_flag(new_flag=True)
        self.matrix_update()

    def dual_update_step(self):
        for neigh in self.agent.out_neighbors:
            self.nu_to[neigh] += float(self.rho_to[neigh]) * (self.x - self.x_neigh[neigh])
            if np.linalg.norm(self.x - self.x_neigh[neigh]) >= self.gamma * self.gr_cn[neigh]:
                self.rho_to[neigh] = min(self.beta * self.rho_to[neigh], self.max_penalty)
            # norm of the constraint value
            self.gr_cn[neigh] = np.linalg.norm(self.x-self.x_neigh[neigh])

        for idx, constr in enumerate(self.agent.problem.constraints):
            if constr.is_equality:
                self.multiplier[idx] += self.penalty[idx] * constr.function.eval(self.x)
                if np.linalg.norm(constr.function.eval(self.x), ord=2) > self.gamma * self.constr_val[idx]:
                    self.penalty[idx] = min(self.beta * self.penalty[idx], self.max_penalty)
                self.constr_val[idx] = np.linalg.norm(constr.function.eval(self.x), ord=2)
            elif constr.is_inequality:
                self.multiplier[idx] = np.max(
                    [np.zeros(constr.output_shape),
                     self.multiplier[idx] + self.penalty[idx] * constr.function.eval(self.x)],
                    axis=0)
                if np.linalg.norm(
                    np.max([-self.multiplier[idx] / self.penalty[idx],
                            constr.function.eval(self.x)],
                           axis=0)) > self.gamma * self.constr_val[idx]:
                    self.penalty[idx] = min(self.beta * self.penalty[idx], self.max_penalty)
                self.constr_val[idx] = np.linalg.norm(
                    np.max([-self.multiplier[idx]/self.penalty[idx], constr.function.eval(self.x)], axis=0))

        self.M_done = True

    def reset_step(self):
        """
        Reset the matrix S and update e
        """
        self.M_done = False
        self.matrix_reset()
        self.e = self.e/1.5
        self.__update_stepsize()

    def run(self, running_time: float = 10.0):
        if not isinstance(running_time, float):
            raise TypeError("Running time must be a float")
        if self.enable_log:
            dims = [1]
            for dim in self.x.shape:
                dims.append(dim)
            self.dims = dims
            self.sequence = np.zeros(dims)
            self.sequence[0] += self.x
            self.timestamp_sequence_awake = [time.time()]
            self.timestamp_sequence_sleep = [time.time()]

        # Exchange all data at the beginning
        data = self.agent.neighbors_exchange(self.x)
        for neigh in data:
            self.x_neigh[neigh] = data[neigh]
        self.__initialize_multipliers()
        self.__update_local_lagrangian()
        self.__update_stepsize()

        received_nu = 0
        end_time = time.time() + running_time
        while time.time() < end_time:
            if self.enable_log:
                self.timestamp_sequence_awake.append(time.time())

            data = self.agent.neighbors_receive_asynchronous()
            for neigh in data:
                msg = data[neigh]
                if msg["type"] == "primal":
                    self.x_neigh[neigh] = msg['x']
                    self.update_column(neigh, msg['S'])
                elif msg["type"] == "dual":
                    self.force_matrix_update()
                    self.nu_from[neigh] = msg['nu']
                    self.rho_from[neigh] = msg['rho']
                    received_nu += 1
                del(msg)

            if self.M_done and received_nu == len(self.agent.in_neighbors):
                received_nu = 0
                self.reset_step()
                self.__update_local_lagrangian()

            if not self.check_stop() and not self.M_done:
                self.primal_update_step()
                msg = {'type': "primal",
                       'x': self.x,
                       'S': self.S[:, -1]}
                self.agent.neighbors_send(msg)

            if self.check_stop() and not self.M_done:
                self.dual_update_step()
                for neigh in self.agent.out_neighbors:
                    msg = {'type': "dual",
                           'nu': self.nu_to[neigh],
                           'rho': self.rho_to[neigh]}
                    self.agent.communicator.neighbors_send(msg, [neigh])

            # save sequence
            if self.enable_log:
                self.sequence = np.vstack([self.sequence, self.x.reshape(self.dims)])

            if self.enable_log:
                self.timestamp_sequence_sleep.append(time.time())

        if self.enable_log:
            return self.timestamp_sequence_awake, \
                self.timestamp_sequence_sleep, \
                self.sequence

    def get_result(self):
        """Return the value of the solution
        """
        return self.x



from .problem import Problem
from ..functions import AbstractFunction, AffineForm
from .milp import MixedIntegerLinearProblem


class ConstraintCoupledProblem(Problem):
    """A local part of a constraint-coupled problem. 

    Args:
        objective_function (AbstractFunction, optional): Local objective function. Defaults to None.
        constraints (list, optional): Local constraints. Defaults to None.
        coupling_function (AbstractFunction, optional): Local function contributing to coupling constraints. Defaults to None.

    Attributes:
        objective_function (Function): Objective function to be minimized
        constraints (AbstractSet or Constraint): Local constraints
        coupling_function (Function): Local function contributing to coupling constraints
    """

    coupling_function = None

    def __new__(cls, *args, **kwargs):
        instance = object.__new__(cls)
        return instance

    def __init__(self, objective_function: AbstractFunction = None, constraints: list = None,
                 coupling_function: AbstractFunction = None, **kwargs):
        super().__init__(objective_function=objective_function, constraints=constraints, **kwargs)
        if coupling_function is not None:
            self.set_coupling_function(coupling_function)

    def set_coupling_function(self, fn: AbstractFunction):
        """Set the coupling constraint function

        Args:
            fn: coupling constraint function

        Raises:
            TypeError: input must be a AbstractFunction 
        """
        if not isinstance(fn, AbstractFunction):
            raise TypeError("coupling function must be a AbstractFunction")
        self.coupling_function = fn


class ConstraintCoupledMILP(ConstraintCoupledProblem, MixedIntegerLinearProblem):

    def __init__(self, objective_function: AffineForm, coupling_function: AffineForm, constraints: list = None, integer_vars: list = None, binary_vars: list = None, **kwargs):
        """[summary]

        Args:
            objective_function (AffineForm): [description]
            coupling_function (AffineForm): Local function contributing to coupling constraints
            constraints (list, optional): [description]. Defaults to None.
            integer_vars (list, optional)
            binary_vars (list, optional)
        """

        if not coupling_function.is_affine:
            raise TypeError("Coupling function must be affine")
        super().__init__(objective_function=objective_function, coupling_function=coupling_function, constraints=constraints, integer_vars=integer_vars, binary_vars=binary_vars)



import numpy as np
from typing import Union, Callable
from ..agents.agent import Agent
from .consensus import Consensus
from ..problems.problem import Problem


class DualSubgradientMethod(Consensus):
    """Distributed dual subgradient method.

    From the perspective of agent :math:`i` the algorithm works as follows. For :math:`k=0,1,\\dots`

    .. math::

        y_i^{k} &= \sum_{j=1}^N w_{ij} \lambda_j^k

        x_i^{k+1} &\in \\text{arg} \min_{x_i \in X_i} f_i(x_i) + g_i(x_i)^\\top y_i^k

        \lambda_i^{k+1} &= \Pi_{\lambda \ge 0}[y_i^k + \\alpha^k g_i(x_i^{k+1})]

        \hat{x}_i^{k+1} &= \hat{x}_i^{k} + \\frac{\\alpha^k}{\sum_{r=0}^{k} \\alpha^r} (x_i^{k+1} - \hat{x}_i^{k})

    where :math:`x_i, \hat{x}_i\\in\\mathbb{R}^{n_i}`, :math:`\lambda_i,y_i\\in\\mathbb{R}^S`, :math:`X_i\\subseteq\mathbb{R}^{n_i}` for all :math:`i`, :math:`\\alpha^k` is a positive stepsize and :math:`\\Pi_{\lambda \ge 0}[]` denotes the projection operator over the nonnegative orthant.

    The algorithm has been presented in [FaMa17]_.

    .. warning::
        this algorithm is still under development
    """

    def __init__(self, agent: Agent, initial_condition: np.ndarray, initial_runavg: np.ndarray = None, enable_log: bool = False):
        super(DualSubgradientMethod, self).__init__(agent, initial_condition, enable_log)

        # initalize running average
        if initial_runavg is None:
            self.x_hat = np.zeros(self.agent.problem.objective_function.input_shape)
        else:
            self.x_hat = initial_runavg

        # initialize cumulative sum of stepsize
        self.stepsize_sum = 0
        
        # number of coupling constraints
        self.S = initial_condition.size

    def _update_local_solution(self, x: np.ndarray, stepsize: float = 0.1):
        y = x

        # define lagrangian: f_i(x_i) + y^top g_i(x_i)
        lagrangian = self.agent.problem.objective_function + (y @ self.agent.problem.coupling_function)

        # create local problem
        pb = Problem(lagrangian, self.agent.problem.constraints)

        # solve local problem
        x_lagr = pb.solve()

        # perform a sugradient step on the dual problem
        lambda_t = y + stepsize * self.agent.problem.coupling_function.eval(x_lagr)

        # project onto non-negative orthant
        lambda_t = np.maximum(lambda_t, np.zeros((self.S, 1)))
        self.x = lambda_t

        # update running average
        self.stepsize_sum += stepsize
        self.x_hat = self.x_hat + stepsize/self.stepsize_sum * (x_lagr - self.x_hat)


    def run(self, iterations: int = 1000, stepsize: Union[float, Callable] = 0.1, verbose: bool=False,
        callback_iter: Callable=None) -> np.ndarray:
        """Run the algorithm for a given number of iterations

        Args:
            iterations: Number of iterations. Defaults to 1000.
            stepsize: If a float is given as input, the stepsize is constant. 
                                                         If a function is given, it must take an iteration k as input and output the corresponding stepsize.. Defaults to 0.1.
            verbose: If True print some information during the evolution of the algorithm. Defaults to False.
            callback_iter: callback function to be called at the end of each iteration. Must take an iteration k as input. Defaults to None.

        Raises:
            TypeError: The number of iterations must be an int
            TypeError: The stepsize must be a float or a callable

        Returns:
            return a tuple (lambda, x_hat) with the sequence of dual and primal estimates if enable_log=True.
        """
        if not isinstance(iterations, int):
            raise TypeError("The number of iterations must be an int")
        if not (isinstance(stepsize, float) or callable(stepsize)):
            raise TypeError("The stepsize must be a float or a function")
        if callback_iter is not None and not callable(callback_iter):
            raise TypeError("The callback function must be a Callable")

        if self.enable_log:
            # initialize sequence of lambda
            lambda_dims = [iterations]
            for dim in self.x.shape:
                lambda_dims.append(dim)
            self.lambda_sequence = np.zeros(lambda_dims)

            # initialize sequence of x hat
            xhat_dims = [iterations]
            for dim in self.x_hat.shape:
                xhat_dims.append(dim)
            self.xhat_sequence = np.zeros(xhat_dims)

        for k in range(iterations):
            if not isinstance(stepsize, float):
                step = stepsize(k)
            else:
                step = stepsize
            self.iterate_run(stepsize=step)

            if callback_iter is not None:
                callback_iter(k)

            if self.enable_log:
                self.lambda_sequence[k] = self.x
                self.xhat_sequence[k] = self.x_hat
            
            if verbose:
                if self.agent.id == 0:
                    print('Iteration {}'.format(k), end="\r")

        if self.enable_log:
            return (self.lambda_sequence, self.xhat_sequence)
    
    def get_result(self):
        """Return the actual value of dual and primal averaged variable
    
        Returns:
            tuple (dual, primal) of numpy.ndarray: value of primal running average, value of dual variable
        """
        return (self.x, self.x_hat)


import numpy as np
from typing import Union
from .constraints import Constraint
from copy import deepcopy


class ExtendedConstraint(Constraint):
    """Constraint with extended variable

    Args:
        constr (Constraint or list of Constraint): original constraint(s)
        n_var: number of additional variables. Defaults to 1
        axis: axis along which the additional variables are appended. Defaults to -1 (the last valid one)
        pos: position index of the old variable vector. Defaults to 0

    Raises:
        TypeError: fn must be a Constraint object or a list of Constraint objects
        TypeError: n_var must be a positive int
        TypeError: axis must be int
    """

    def __new__(cls, constr: Union[Constraint, list], n_var: int = 1, axis=-1, pos=0):
        if not isinstance(constr, Constraint) and not isinstance(constr, list):
            raise TypeError("fn must be a Constraint object or a list of Constraint objects")

        if isinstance(constr, list) and not all(isinstance(x, Constraint) for x in constr):
            raise TypeError("fn must be a Constraint object or a list of Constraint objects")

        if not isinstance(n_var, int) or n_var <= 0:
            raise TypeError("n_var must be a positive int")

        if not isinstance(axis, int):
            raise TypeError("axis must be int")

        if isinstance(constr, Constraint):
            from ..functions.extended_function import ExtendedFunction
            return Constraint(ExtendedFunction(constr.fn, n_var, axis, pos), constr.sign)
        else:
            return [ExtendedConstraint(x, n_var, axis, pos) for x in constr]



import numpy as np
import warnings
from typing import Union
from .abstract_function import AbstractFunction
from .utilities import check_input
from ..utils.utilities import is_pos_def, is_semi_pos_def


class QuadraticForm(AbstractFunction):
    """Quadratic form

    .. math::

        f(x)= x^\\top P x + q^\\top x + r

    with :math:`P\\in \\mathbb{R}^{n\\times n}`, :math:`q\\in \\mathbb{R}^{n}`, :math:`r\\in \\mathbb{R}` and :math:`x: \\mathbb{R}^{n}`.

    Args:
        fn (AbstractFunction): input function
        P (numpy.ndarray, optional): input matrix. Defaults to None (identity).
        q (numpy.ndarray, optional): input vector. Defaults to None (zero).
        r (numpy.ndarray, optional): input bias. Defaults to None (zero).

    Raises:
        TypeError: First argument must be a AbstractFunction object
        TypeError: Second argument must be a numpy.ndarray
        ValueError: Input matrix must be a square matrix
        ValueError: Dimension mismatch. Input matrix must have shape compliant with the function output shape
    """

    def __init__(self, fn: AbstractFunction, P: np.ndarray = None, q: np.ndarray = None, r: np.ndarray = None):
        if not isinstance(fn, AbstractFunction):
            raise TypeError("First argument must be a AbstractFunction object")

        if not fn.is_differentiable:
            warnings.warn(
                'Composition with a nondifferentiable function will \
                    lead to an error when asking for a subgradient')
        else:
            self.differentiable = True

        if P is not None:
            if not isinstance(P, np.ndarray):
                raise TypeError("Second argument must be a numpy.ndarray")
            row, col = P.shape
            if row != col:
                raise ValueError("Input matrix must be a square matrix")
            if P.shape[0] != fn.output_shape[0]:
                raise ValueError(
                    "Dimension mismatch. Input matrix must have shape {}".format(
                        (fn.output_shape[0], fn.output_shape[0])))
            # if not is_semi_pos_def(P):
            #     warnings.warn("Warning, input matrix is not (semi)positive definite")
        else:
            P = np.eye(fn.output_shape[0])

        if q is not None:
            if not isinstance(q, np.ndarray):
                raise TypeError("third argument must be a numpy.ndarray")

            if q.shape != fn.output_shape:
                raise ValueError(
                    "Dimension mismatch. Input vector must have shape {}".format(fn.output_shape))
        else:
            q = np.zeros(fn.output_shape)

        if r is not None:
            if not isinstance(q, (np.ndarray, float, int)):
                raise TypeError("4-th argument must be a int, float or numpy.ndarray")

            if isinstance(r, np.ndarray):
                if r.shape != (1, 1):
                    raise ValueError(
                        "Dimension mismatch. Input bial must have shape (1,1)")
        else:
            r = 0.0

        self.input_shape = fn.input_shape
        self.output_shape = (1, 1)

        if fn.is_affine:
            self.quadratic = True
            # (A^T x + b)^T P (A^T x + b) + q^T(A^T x + b) + r =
            # = x^T A P A^T x + x^T A P b + b^T P A^T x + b^T P b + q^T A^T x + q^T b + r
            # = x^T A P A^T x + (b^T(P + P^T)A^T + q^t A^t) x + b^T P b + q^T b + r
            #       -------     ---------------------------     -------------------
            # = x^T   P     x +             q^t             x +          r
            from .variable import Variable
            A = fn.A
            b = fn.b
            self.P = A @ P @ A.transpose()
            self.q = (b.transpose() @ (P + P.transpose()) @ A.transpose() +
                      q.transpose() @ A.transpose()).transpose()
            self.r = b.transpose() @ P @ b + q.transpose() @ b + r
            self.fn = Variable(self.input_shape[0])

        else:
            self.quadratic = False
            self.fn = fn
            self.P = P
            self.q = q
            self.r = r

        self.affine = False

        super().__init__()

    def _expression(self):
        expression = 'QuadraticForm({}, P, q, r)'.format(self.fn._expression())
        return expression 
    
    def _to_cvxpy(self):
        import cvxpy as cvx
        fn = self.fn._to_cvxpy()
        return cvx.quad_form(fn, self.P) + self.q.transpose() @ fn + self.r.flatten()
    
    def _extend_variable(self, n_var, axis, pos):
        return QuadraticForm(self.fn._extend_variable(n_var, axis, pos), self.P, self.q, self.r)

    def get_parameters(self):
        if self.is_quadratic:
            return self.P, self.q, self.r

    def __neg__(self):  # TODO Check
        P = -1 * self.P
        q = -1 * self.q
        r = -1 * self.r
        return QuadraticForm(self.fn, P, q, r)

    def __add__(self, other):
        if isinstance(other, (int, float, np.ndarray)):
            P = self.P
            q = self.q
            r = self.r + other
            return QuadraticForm(self.fn, P, q, r)
        elif isinstance(other, AbstractFunction):
            if self.is_quadratic and other.is_affine:
                other_A, other_b = other.get_parameters()
                if self.q.shape == other_A.shape:
                    P = self.P
                    q = self.q + other_A
                    r = self.r + other_b
                    return QuadraticForm(self.fn, P, q, r)
                else:
                    raise ValueError("Incompatible dimensions")
            if self.is_quadratic and other.is_quadratic:
                other_P, other_q, other_r = other.get_parameters()
                if self.P.shape == other_P.shape:
                    P = self.P + other_P
                    q = self.q + other_q
                    r = self.r + other_r
                    return QuadraticForm(self.fn, P, q, r)
                else:
                    raise ValueError("Incompatible dimensions")
            else:
                return super().__add__(other)
        else:
            raise TypeError

    def __radd__(self, other):
        return self.__add__(other)

    def __iadd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        if isinstance(other, (int, float, np.ndarray)):
            P = self.P
            q = self.q
            r = self.r - other
            return QuadraticForm(self.fn, P, q, r)
        elif isinstance(other, AbstractFunction):
            if self.is_quadratic and other.is_affine:
                other_A, other_b = other.get_parameters()
                if self.q.shape == other_A.shape:
                    P = self.P
                    q = self.q - other_A
                    r = self.r - other_b
                    return QuadraticForm(self.fn, P, q, r)
                else:
                    raise ValueError("Incompatible dimensions")
            if self.is_quadratic and other.is_quadratic:
                other_P, other_q, other_r = other.get_parameters()
                if self.P.shape == other_P.shape:
                    P = self.P - other_P
                    q = self.q - other_q
                    r = self.r - other_r
                    return QuadraticForm(self.fn, P, q, r)
                else:
                    raise ValueError("Incompatible dimensions")
            else:
                return super().__sub__(other)
        else:
            raise TypeError

    def __rsub__(self, other):
        return self.__neg__().__add__(other)

    def __isub__(self, other):
        return self.__sub__(other)

    def __mul__(self, other):
        if isinstance(other, (int, float)):
            P = other * self.P
            q = other * self.q
            r = other * self.r
            return QuadraticForm(self.fn, P, q, r)
        else:
            return super().__mul__(other)

    def __rmul__(self, other: Union[int, float]):
        return self.__mul__(other)

    def __imul__(self, other):
        return self.__mul__(other)

    @check_input
    def eval(self, x: np.ndarray) -> np.ndarray:
        p1 = self.fn.eval(x).transpose() @ self.P @ self.fn.eval(x)
        p2 = self.q.transpose() @ self.fn.eval(x)
        p3 = self.r
        return (p1 + p2 + p3).reshape(self.output_shape)

    @check_input
    def _alternative_jacobian(self, x: np.ndarray, **kwargs) -> np.ndarray:
        if not self.fn.is_differentiable:
            warnings.warn("Composition of non affine functions. The Jacobian may be not correct.")
        return self.fn.eval(x).transpose().dot((self.P +
                                                self.P.transpose()).dot(
            self.fn.jacobian(x, **kwargs)
        )) + self.q.transpose().dot(self.fn.jacobian(x, **kwargs))

    # @check_input
    # def hessian(self, x: np.ndarray = None, **kwargs) -> np.ndarray:
    #     from .variable import Variable
    #     if isinstance(self.fn, Variable):
    #         return self.P
    #     else:
    #         return NotImplemented



import numpy as np
from ..agents import Agent
from .algorithm import Algorithm
from ..problems import Problem, LinearProblem
from ..functions import QuadraticForm, Variable
from itertools import combinations
from typing import Tuple
from threading import Event

# TODO simplify code and add more detailed comments to methods
class ConstraintsConsensus(Algorithm):
    """Constraints Consensus Algorithm [NoBu11]_
    
    This algorithm solves convex and abstract programs in the form

        .. math::
            :nowrap:

            \\begin{split}
            \min_{x} \: & \: c^\\top x \\\\
            \mathrm{subj. to} \: & \: x\\in \\bigcap_{i=1}^{N} X_i \\\\
            & \: x \ge 0
            \\end{split}

    Attributes:
        x (numpy.ndarray): current value of the local solution
        B (numpy.ndarray): basis associated to the local solution
        shape (tuple): shape of the variable
        x_neigh (dict): dictionary containing the local solution of the (in-)neighbors
        sequence_x (numpy.ndarray): sequence of local solutions

    Args:
        agent (Agent): agent to execute the algorithm
        enable_log (bool): True to enable log
    """

    def __init__(self, agent: Agent, enable_log: bool = False):

        # initialize variables
        self.agent = agent
        self.enable_log = enable_log
        self.sequence_x = None
        self.shape = self.agent.problem.input_shape
        self.x_neigh = {}

        # initialize optimization variable and basis
        self.x = self.agent.problem.solve()
        self.B = self.compute_basis(self.agent.problem.constraints)

    def get_basis(self):
        """Return agent basis
        """
        return self.B

    def compute_basis(self, constraints):
        """Compute a (minimal) basis for the given constraint list
        """

        # remove redundant constraints
        constraints_unique = self.unique_constr(constraints)

        # find active constraints
        active_constr = []
        for con in constraints_unique:
            # try:
            val = con.function.eval(self.x)
            # except:
            if abs(val) < 1e-5:
                active_constr.append(con)

        # enumerating the possible combinations with d+1 constraints
        basis = active_constr
        for cand_basis_tuple in combinations(active_constr, np.max(self.shape) + 1):
            # convert candidate basis to list
            cand_basis = list(cand_basis_tuple)

            # solve problem with this candidate basis
            prob = Problem(self.agent.problem.objective_function, cand_basis)
            
            try:
                solution = prob.solve()
                if np.array_equal(solution, self.x):
                    # we have found a basis, break cycle
                    basis = cand_basis
                    break
            except:
                pass
        return basis

    def unique_constr(self, constraints):
        """Remove redundant constraints from given constraint list
        """

        # initialize shrunk list of constraints
        con_shrink = []

        # cycle over given constraints
        for con in constraints:
            if not con_shrink:
                con_shrink.append(con)
            else:
                check_equal = np.zeros((len(con_shrink), 1))
                # TODO: use numpy array_equal

                # cycle over already added constraints
                for idx, con_y in enumerate(con_shrink):
                    if self.compare_constr(con_y, con):
                        check_equal[idx] = 1
                n_zero = np.count_nonzero(check_equal)

                # add constraint if different from all the others
                if n_zero == 0:
                    con_shrink.append(con)
        return con_shrink

    def compare_constr(self, a, b):
        """Compare two constraints to check whether they are equal
        """

        # test for affine constraints
        if (a.is_affine and b.is_affine):
            A_a, b_a = a.get_parameters()
            A_b, b_b = b.get_parameters()
            if np.array_equal(A_a, A_b) and np.array_equal(b_a, b_b):
                return True
            else:
                return False
        
        # test for quadratic constraints
        elif (a.is_quadratic and b.is_quadratic):
            P_a, q_a, r_a = a.get_parameters()
            P_b, q_b, r_b = b.get_parameters()
            if np.array_equal(P_a, P_b) and np.array_equal(q_a, q_b) and np.array_equal(r_a, r_b):
                return True
            else:
                return False
        else:
            return False

    def constr_to_dict(self, constraints):
        """Convert constraint list to dictionary
        """

        con_dict = {}
        for key, con in enumerate(constraints):
            if con.is_affine:
                A, b = con.get_parameters()
                con_dict[key] = {"kind": "affine", "sign": con.sign, "Amat": A,
                                 "bmat": b}
            if con.is_quadratic:
                P, q, r = con.get_parameters()
                con_dict[key] = {"kind": "quadratic", "sign": con.sign, "Pmat": P,
                                 "qmat": q, "rmat": r}
        return con_dict

    def dict_to_constr(self, dictio):
        """Convert dictionary to constraint list
        """

        dict_len = len(dictio)
        constr = []
        x = Variable(int(np.max(self.shape)))
        for idx in np.arange(dict_len):
            con_dict = dictio.get(idx)
            str_kind = con_dict.get("kind")
            str_sign = con_dict.get("sign")
            if str_kind == "quadratic":
                P = con_dict.get("Amat")
                q = con_dict.get("qmat")
                r = con_dict.get("rmat")
                g = QuadraticForm(x, P, q, r)
                if str_sign == "==":
                    con = g == 0
                else:
                    con = g <= 0
            elif str_kind == "affine":
                A = con_dict.get("Amat")
                b = con_dict.get("bmat")
                if str_sign == "==":
                    con = A @ x + b == 0
                else:
                    con = A @ x + b <= 0
            else:
                raise ValueError(
                    "Current supported constraints are affine and quadratic")
            constr.append(con)
        return constr

    def iterate_run(self, event: Event):
        """Run a single iterate of the algorithm
        """

        # convert basis to dictionary and send to neighbors
        basis_dict = self.constr_to_dict(self.B)
        data = self.agent.neighbors_exchange(basis_dict, event=event)

        if event is not None and event.is_set():
            return

        # create list of agent's constraints + received constraints
        constraints = []
        constraints.extend(self.agent.problem.constraints)
        for neigh in data:
            constraints.extend(self.dict_to_constr(data[neigh]))

        # solve problem
        problem = Problem(self.agent.problem.objective_function, constraints)
        self.x = problem.solve()

        # compute new basis
        self.B = self.compute_basis(constraints)

    def run(self, iterations=100, verbose: bool=False, event: Event=None, **kwargs) -> np.ndarray:
        """Run the algorithm for a given number of iterations

        Args:
            iterations (int, optional): Number of iterations. Defaults to 100.
            verbose: If True print some information during the evolution of
                    the algorithm. Defaults to False.

        Returns:
            the sequence of computed solutions if enable_log=True.
        """
        if self.enable_log:
            dims = [iterations]
            for dim in self.shape:
                dims.append(dim)
            self.sequence_x = np.zeros(dims)
        
        last_iter = np.copy(iterations)

        for k in range(iterations):
            self.iterate_run(event)

            # check if we must stop for external event
            if event is not None and event.is_set():
                last_iter = k
                break

            if self.enable_log:
                self.sequence_x[k] = self.x

            if verbose:
                if self.agent.id == 0:
                    print('Iteration {}'.format(k), end="\r")

        if self.enable_log:
            return self.sequence_x.take(np.arange(0,last_iter), axis=0)

    def get_result(self):
        """Return the current value of x

        Returns:
            numpy.ndarray: value of x
        """
        return self.x

class DistributedSimplex(Algorithm):
    """Distributed Simplex Algorithm [BuNo11]_

        This algorithm solves linear programs in standard form

        .. math::
            :nowrap:

            \\begin{split}
            \min_{x} \: & \: c^\\top x \\\\
            \mathrm{subj. to} \: & \: Ax = b \\\\
            & \: x \ge 0
            \\end{split}
        
        When reading the variable agent.problem.constraints, this class only
        considers equality constraints. Other constraints are discarded.

    Attributes:
        x (numpy.ndarray): current value of the complete solution
        J (float): current value of the cost
        x_basic (numpy.ndarray): current value of the basic solution
        B (numpy.ndarray): basis associated to the local solution
        n_constr (tuple): number of constraints of the problem
        B_neigh (dict): dictionary containing the local bases of (in-)neighbors
        A_init (numpy.ndarray): initial constraint matrix
        b_init (numpy.ndarray): initial constraint vector
        c_init (numpy.ndarray): initial cost vector
        sequence_x (numpy.ndarray): sequence of solutions
        sequence_J (numpy.ndarray): sequence of costs

    Args:
        agent (Agent): agent to execute the algorithm
        problem_size (list): total number of variables in the network. Defaults to None.
            If both problem_size and local_indices is provided, the complete solution
            vector will be computed.
        local_indices (list): indices of the agent's variables in the network, starting from 0. Defaults to None.
            If both problem_size and local_indices is provided, the complete solution
            vector will be computed.
        enable_log (bool): True to enable log
        stop_iterations (int): iterations with constant solution to stop algorithm. Defaults to None (disabled).
        big_M(float): cost of big-M variables. Defaults to 500.
    """

    def __init__(self, agent: Agent, problem_size: int = None, local_indices: list = None,
                enable_log: bool = False, stop_iterations: int = None, big_M: float = 500.0):

        # initialize class variables
        self.agent = agent
        self.problem_size = problem_size
        self.local_indices = local_indices
        self.enable_log = enable_log
        self.sequence_x = None
        self.sequence_J = None
        self.stop_iterations = stop_iterations
        self.big_M = big_M
        self.B_neigh = {}

        # simplex algorithm parameters
        self.par_iters = 100
        self.par_tol_leav = 1e-6
        self.par_tol_ent  = 1e-6

        # check if input data is correct
        if not isinstance(self.agent.problem, LinearProblem):
            raise TypeError("agent.problem must be an instance of LinearProblem")

        # other variables
        self.read_problem_data()
        self.n_constr = self.A_init.shape[0]
        self.x_basic = None # primal solution (basic)
        self.J = None # cost
        self.x_dual = None # dual solution
        self.x = None # primal solution (full)

        self.save_x_dual = True # true if self.x_dual can be populated
        self.save_x = False # true if self.x can be populated
        self.indices_available = False # true if column indices are available

        # check index consistency and, in case, adjust flags
        if self.problem_size is not None and self.local_indices is not None:
            self.check_index_consistency()
            self.save_x = True
            self.indices_available = True
        
        # initialize algorithm
        self.initialize()
    
    def check_index_consistency(self):
        """Check consistency of local indices, problem_size and constraint matrix
        """
        if not isinstance(self.local_indices, list):
            raise TypeError("local_indices must be a list")
        if len(self.local_indices) != self.A_init.shape[1]:
            raise ValueError("The size of local_indices must equal the number of variables in agent.problem")
        if self.problem_size < self.A_init.shape[1]:
            raise ValueError("problem_size must be greater than or equal to the size of local_indices")
    
    def read_problem_data(self):
        """Read local problem data from agent.problem. The data is
        saved in order to be solved as a standard form problem.
        """
        A_list = []
        b_list = []

        # loop through equality constraints
        for constraint in self.agent.problem.constraints:
            if constraint.is_equality:
                A_c, b_c = constraint.get_parameters()

                # constraint is written in the form A^\top x + b = 0
                A_list.append(A_c.transpose())
                b_list.append(-b_c.flatten()) # save as 1D vector
        
        # build initial constraint matrix and vector from lists
        self.b_init = np.hstack(tuple(b_list))[:, None] # save as 2D vertical vector
        self.A_init = np.vstack(tuple(A_list))

        # read initial cost vector
        cost_parameters = self.agent.problem.objective_function.get_parameters()
        self.c_init = cost_parameters[0] # save as 2D vertical vector

    def initialize(self):
        """Evaluate a first solution and basis starting from
        agent's constraints through the Big-M method
        """

        # change sign of constraint rows where b is negative
        for j in range(self.b_init.shape[0]):
            if self.b_init[j] < 0:
                self.b_init[j] *= -1
                self.A_init[j, :] *= -1
        
        # tableau for agent's initial matrices
        H_init = np.vstack((self.c_init.transpose(), self.A_init))

        # tableau for big-M (initial lex-feasible basis)
        A_bigM = np.eye(self.n_constr)
        c_bigM = self.big_M*np.ones((1, self.n_constr))
        H_bigM = np.vstack((c_bigM, A_bigM))

        # add column partitioning information if available
        if self.indices_available:
            # initial columns are assigned local_indices
            H_init = np.vstack((H_init, np.array(self.local_indices)))

            # big-M columns are assigned indices after problem_size
            indices_bigM = np.arange(self.problem_size, self.problem_size + self.n_constr)
            H_bigM = np.vstack((H_bigM, np.array(indices_bigM)))

        # prepare lex-sorted tableau
        H_sort, basis_idx = sort_tableau(H_init, H_bigM)

        if self.indices_available:
            tableau = H_sort[0:-1, :]
        else:
            tableau = H_sort

        # run simplex method
        sol = lex_simplex(tableau, self.b_init, basis_idx, self.par_iters, self.par_tol_ent, self.par_tol_leav)

        # update local solution
        self._update_local_solution(H_sort[:, sol[1]])

    def _update_local_solution(self, basis):
        """Save solution data using current basis information
        """
        # extract constraint matrix and cost vector
        if self.indices_available:
            A = basis[1:-1, :]
        else:
            A = basis[1:, :]
        c = basis[0, :]

        # compute basic variables
        x_basic = np.linalg.solve(A, self.b_init)

        # compute dual solution
        y = np.linalg.solve(A.transpose(), c)

        # save
        self.B = basis
        self.x_basic = x_basic
        self.x_dual = y
        self.J = c @ x_basic

        # compute entire vector if column partitioning information is available
        if self.indices_available:
            self.x = np.zeros((self.problem_size + self.n_constr, 1))
            basis_idx = basis[-1, :].astype(int)
            self.x[basis_idx, :] = self.x_basic
    
    def get_basis(self):
        """Return current basis
        """
        if self.indices_available:
            return self.B[0:-1]
        else:
            return self.B
    
    def get_result(self):
        """Return the current value of the solution
    
        Returns:
            tuple of nd.ndarray (primal, primal_basic, dual, cost): value of primal solution, primal basic solution, dual solution, cost
        """
        return (self.x, self.x_basic, self.x_dual, self.J)

    def iterate_run(self, event: Event):
        """Run a single iterate of the algorithm
        """

        # exchange columns with neighbors - TODO implement exchange of "null" symbol
        data = self.agent.neighbors_exchange(self.B, event=event)

        if event is not None and event.is_set():
            return

        # concatenate all the received bases
        for neigh in data:
            self.B_neigh[neigh] = data[neigh]
        H_list = list(self.B_neigh.values())
        
        # add tableau for agent's initial matrices
        if self.indices_available:
            H_list.append(np.vstack((self.c_init.transpose(), self.A_init, self.local_indices)))
        else:
            H_list.append(np.vstack((self.c_init.transpose(), self.A_init)))

        # prepare lex-sorted tableau
        H_full = np.hstack(tuple(H_list))
        H_sort, basis_idx = sort_tableau(H_full, self.B)

        if self.indices_available:
            tableau = H_sort[0:-1, :]
        else:
            tableau = H_sort

        # run simplex method
        sol = lex_simplex(tableau, self.b_init, basis_idx, self.par_iters, self.par_tol_ent, self.par_tol_leav)

        # update local solution
        self._update_local_solution(H_sort[:, sol[1]])

    def run(self, iterations=100, verbose: bool=False, event: Event=None, **kwargs) -> np.ndarray:
        """Run the algorithm for a given number of iterations

        Args:
            iterations (int, optional): Maximum number of iterations. Defaults to 100.
            verbose: If True print some information during the evolution of the algorithm. Defaults to False.

        Raises:
            TypeError: The number of iterations must be an int
        
        Returns:
            return a tuple (x, J) with the sequence of solutions and costs if enable_log=True.
        """
        if not isinstance(iterations, int):
            raise TypeError("The number of iterations must be an int")

        if self.enable_log:
            # initialize cost sequence
            dims = [iterations]
            self.sequence_J = np.zeros(dims)

            # initialize solution sequence
            if self.save_x:
                dims.extend(self.x.shape)
                self.sequence_x = np.zeros(dims)
        
        # initialize counter for stopping criterion
        counter = 0
        last_iter = np.copy(iterations)

        for k in range(iterations):
            # store previous basis
            prev_B = self.B

            # perform an iteration
            self.iterate_run(event)

            # check if we must stop for external event
            if event is not None and event.is_set():
                last_iter = k
                break

            # store solution and cost sequence
            if self.enable_log:
                self.sequence_J[k] = self.J

                if self.save_x:
                    self.sequence_x[k] = self.x
            
            # print information
            if verbose and self.agent.id == 0:
                print('Iteration {}'.format(k), end="\r")

            # increase counter if solution has not changed, otherwise reset
            if np.linalg.norm(prev_B - self.B) < 1e-6:
                counter += 1
            else:
                counter = 0
            
            # check termination condition
            if self.stop_iterations is not None and counter > self.stop_iterations:
                # convergence detected
                self.agent.neighbors_send(self.B) # broadcast local basis a last time
                last_iter = k+1
                break
        
                # TODO check if problem is infeasible (i.e. if there are still big-M columns in the basis)

        # return sequences
        if self.enable_log:
            if self.save_x:
                return (self.sequence_x.take(np.arange(0,last_iter), axis=0), self.sequence_J[0:last_iter])
            else:
                return (None, self.sequence_J[0:last_iter])


class DualDistributedSimplex(DistributedSimplex):
    """Distributed Simplex Algorithm on dual problem [BuNo11]_

        This algorithm solves linear programs of the form

        .. math::
            :nowrap:

            \\begin{split}
            \max_{x} \: & \: c^\\top x \\\\
            \mathrm{subj. to} \: & \: Ax \le b
            \\end{split}
        
        This class runs the Distributed Simplex algorithm on the (standard form) dual problem.

    Args:
        agent (Agent): agent to execute the algorithm
        num_constraints (list): total number of constraints in the network. Defaults to None.
            If both num_constraints and local_indices is provided, the complete dual solution
            vector will be computed.
        local_indices (list): indices of the agent's constraints in the network, starting from 0. Defaults to None.
            If both num_constraints and local_indices is provided, the complete dual solution
            vector will be computed.
        enable_log (bool): True to enable log
        stop_iterations (int): iterations with constant solution to stop algorithm. Defaults to None (disabled).
        big_M(float): cost of big-M variables. Defaults to 500.
    """

    def __init__(self, agent: Agent, num_constraints: int = None, local_indices: list = None,
                enable_log: bool = False, stop_iterations: int = None, big_M: float = 500.0):
        # call init of parent class
        super().__init__(agent=agent, problem_size=num_constraints, local_indices=local_indices,
                enable_log=enable_log, stop_iterations=stop_iterations, big_M=big_M)
        
        # swap values of save_x and save_x_dual
        self.save_x, self.save_x_dual = self.save_x_dual, self.save_x

    def check_index_consistency(self):
        """Check consistency of local indices, num_constraints and constraint matrix
        """
        # re-adapted error messages for the dual case
        if not isinstance(self.local_indices, list):
            raise TypeError("local_indices must be a list")
        if len(self.local_indices) != self.A_init.shape[1]:
            raise ValueError("The size of local_indices must equal the number of constraints in agent.problem")
        if self.problem_size < self.A_init.shape[1]:
            raise ValueError("num_constraints must be greater than or equal to the size of local_indices")

    def read_problem_data(self):
        """Read local problem data from agent.problem. The data is
        saved in order to be solved as a standard form problem.
        """
        A_list = []
        b_list = []

        # loop through inequality constraints
        for constraint in self.agent.problem.constraints:
            if constraint.is_inequality:
                A_c, b_c = constraint.get_parameters()

                # constraint is written in the form A^\top x + b <= 0
                A_list.append(A_c)
                b_list.append(-b_c.flatten()) # save as 1D vector
        
        # build initial constraint matrix and cost vector from lists
        self.c_init = np.hstack(tuple(b_list))[:, None] # save as 2D vertical vector
        self.A_init = np.vstack(tuple(A_list))

        # read initial constraint vector
        cost_parameters = self.agent.problem.objective_function.get_parameters()
        self.b_init = -cost_parameters[0] # save as 2D vertical vector (minus sign for maximization)

    def _update_local_solution(self, basis):
        """Save solution data using current basis information
        """
        # call method of parent class
        super()._update_local_solution(basis)

        # swap primal and dual solutions
        self.x, self.x_dual = self.x_dual, self.x
    
    def get_result(self):
        """Return the current value of the solution
    
        Returns:
            tuple of nd.ndarray (primal, dual_basic, dual, cost): value of primal solution, dual basic solution, dual solution, cost
        """
        return (self.x, self.x_basic, self.x_dual, self.J)

#####################################
# Utilities

def lexnonpositive(vector: np.ndarray, tol: float) -> bool:
    """tests if a vector is lexicographically non-positive, i.e.,
    it is a vector made of zeros or its first non-zero component is negative

    Args:
        vector (np.ndarray): input vector
        tol (float): numerical tolerance

    Returns:
        is_nonpositive (bool): True if vector is non-positive
    """

    for elem in vector:
        if abs(elem) < tol:  # Treat it as a zero
            continue
        # if the component is positive return false
        if elem > 0:
            return False

        # else return true
        break

    return True


def enteringcolumn(tableau: np.ndarray, ind_basic: np.ndarray, tol: float) -> int:
    """Lex-simplex routine that selects an entering column
    as described in Jones et al (Automatica 2007)

    Args:
        tableau (np.ndarray): lexicographically sorted tableau
        ind_basic (np.ndarray): indices of current lex-feasible basis
        tol (float): numerical tolerance

    Returns:
        index_e (int): index of entering column (or None if basis is already optimal)
    """

    ncols = tableau.shape[1]

    # extract the constraint matrix and the basic columns
    A = tableau[1:, :]
    basis = A[:, ind_basic]

    # evaluate reduced costs and construct matrix [kappa@c kappa] for test
    kappa = np.eye(ncols)
    kappa[:, ind_basic] -= np.linalg.solve(basis, A).transpose()
    cost = tableau[0, :]
    reduced_costs = kappa @ cost
    kc_k_matrix = np.column_stack((reduced_costs, kappa))

    # cycle on non-basic columns
    for ii in range(ncols):
        if ii not in ind_basic:
            test_vector = kc_k_matrix[ii, :]

            if lexnonpositive(test_vector, tol):
                return ii

    return None


def leavingcolumn(index_e: int, b: np.ndarray, tableau: np.ndarray,
    tol: float, ind_basic: np.ndarray) -> int:
    """Lex-simplex routine that selects a leaving column
    with the lex-ratio test described in Jones et al (Automatica 2007)

    Args:
        index_e (int): current entering column
        b (np.ndarray): right-hand side vector of constraints
        tableau (np.ndarray): lexicographically sorted tableau
        tol (float): numerical tolerance
        ind_basic (np.ndarray): indices of current lex-feasible basis

    Returns:
        index_l (int): index of leaving column (or None if problem is unbounded)
    """

    # Extract basic columns from tableau
    basis = tableau[1:, ind_basic]

    # construct beta*A_e (with beta = B^-1)
    A_e = tableau[1:, index_e][:, None]
    beta_A = np.linalg.solve(basis, A_e)

    # find positive entries of beta_A
    ind_pos = np.flatnonzero(beta_A > tol)
    if ind_pos.size == 0:
        return None # problem is unbounded
    
    # construct [beta*b beta] matrix
    beta = np.linalg.inv(basis)
    beta_b = np.linalg.solve(basis, b)
    M = np.column_stack((beta_b, beta))

    # compute ratios
    ratios = M[ind_pos, :] / beta_A[ind_pos]

    # sort rows lexicographically
    ind_sort = np.lexsort(np.rot90(ratios))

    # return lex-minimal row index
    return ind_pos[ind_sort[0]]


def lex_simplex(tableau: np.ndarray, b: np.ndarray, ind_basic: np.ndarray,
    max_iters: int, tol_ent: float, tol_leav: float) -> Tuple[int, np.ndarray, int]:
    """Lexicographic simplex algorithm
    Args:
        tableau (np.ndarray): lexicographically sorted tableau
        b (np.ndarray): right-hand side vector of constraints
        ind_basic (np.ndarray): indices of current lex-feasible basis
        max_iters (int): maximum number of iterations
        tol_ent (float): numerical tolerance to evaluate entering column
        tol_leav (float): numerical tolerance to evaluate leaving column

    Returns:
        unbounded (int): True if problem is unbounded, False otherwise
        new_ind_basic (np.ndarray): new basic variables
        nn (int): number of performed iterations
    """

    new_ind_basic = np.copy(ind_basic)
    
    for nn in range(max_iters):
        # check for an entering column
        index_EC = enteringcolumn(tableau, new_ind_basic, tol_ent)

        if index_EC == None:  # Problem solved
            return False, new_ind_basic, nn+1

        # check for a leaving column
        index_LC = leavingcolumn(index_EC, b, tableau, tol_leav, new_ind_basic)

        if index_LC == None:  # unbounded case
            return True, None, nn+1

        # pivoting
        new_ind_basic[index_LC] = index_EC
        new_ind_basic.sort()

    return False, new_ind_basic, nn+1


def sort_tableau(H_data: np.ndarray, B: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """Lexicographically sort all the columns and remove duplicates
    Args:
        H_data (np.ndarray): tableau without basic columns
        B (np.ndarray): basic columns

    Returns:
        H_sort (np.ndarray): lexicographically sorted tableau without duplicates
        ind_basic_sort (np.ndarray): indices of basic columns
    """
    # make tableau with B at the beginning
    H = np.hstack((B, H_data))
    ind_basic = np.arange(0, B.shape[1]) # basic indices: 0, ..., B.shape[1] - 1

    # remove duplicate columns - TODO use a tolerance
    H_unique, ind_unique = np.unique(H, return_inverse=True, axis=1)

    # lexicographically sort tableau
    ind_sort = np.lexsort(np.flip(H_unique, 0)).tolist()
    H_sort = H_unique[:, ind_sort]

    # compute new position of basic columns
    ind_basic_sort = [ind_sort.index(i) for i in ind_unique[ind_basic]]

    return H_sort, np.array(ind_basic_sort).astype(int)



import numpy as np
import time
import warnings
from copy import deepcopy
from typing import Union, Callable, Tuple
from ..agents import Agent
from ..constraints.projection_sets import AbstractSet
from .consensus import Consensus, BlockConsensus
from disropt.functions.submodular_func import SubmodularFn


class SubgradientMethod(Consensus):
    """Distributed projected (sub)gradient method.

    From the perspective of agent :math:`i` the algorithm works as follows. For :math:`k=0,1,\\dots`

    .. math::

        y_i^{k} &= \sum_{j=1}^N w_{ij} x_j^k

       x_i^{k+1} &= \Pi_{X}[y_i^k - \\alpha^k \\tilde{\\nabla} f_i(y_i^k)]

    where :math:`x_i,y_i\\in\\mathbb{R}^n`, :math:`X\\subseteq\mathbb{R}^n`, :math:`\\alpha^k` is a positive stepsize, :math:`w_{ij}` denotes the weight assigned by agent :math:`i` to agent :math:`j`, :math:`\\Pi_X[]` denotes the projection operator over the set :math:`X` and :math:`\\tilde{\\nabla} f_i(y_i^k)\\in\\partial f_i(y_i^k)` a (sub)gradient of :math:`f_i` computed at :math:`y_i^k`.
    The weight matrix :math:`W=[w_{ij}]_{i,j=1}^N` should be doubly stochastic.

    The algorithm, as written above, was originally presented in [NeOz09]_. Many other variants and extension has been proposed, allowing for stochastic objective functions, time-varying graphs, local stepsize sequences. All these variant can be implemented through the :class:`SubgradientMethod` class.
    """

    def __init__(self, agent: Agent, initial_condition: np.ndarray, enable_log: bool = False):
        if not isinstance(agent, Agent):
            raise TypeError("Agent must be an Agent")
        super(SubgradientMethod, self).__init__(agent, initial_condition, enable_log)

    def _update_local_solution(self, x: np.ndarray, stepsize: float = 0.1, projection: bool = False):
        y = x
        # Perform a sugradient step
        self.x = y - stepsize * self.agent.problem.objective_function.subgradient(y)
        if projection:
            self.x = self.agent.problem.project_on_constraint_set(self.x)

    def run(self, iterations: int = 1000, stepsize: Union[float, Callable] = 0.001, verbose: bool=False) -> np.ndarray:
        """Run the algorithm for a given number of iterations

        Args:
            iterations: Number of iterations. Defaults to 1000.
            stepsize: If a float is given as input, the stepsize is constant. 
                                                         If a function is given, it must take an iteration k as input and output the corresponding stepsize.. Defaults to 0.1.
            verbose: If True print some information during the evolution of the algorithm. Defaults to False.

        Raises:
            TypeError: The number of iterations must be an int
            TypeError: The stepsize must be a float or a callable
            ValueError: Only sets (children of AstractSet) with explicit projections are currently supported
            ValueError: Only one constraint per time is currently supported

        Returns:
            return the sequence of estimates if enable_log=True.
        """
        if not isinstance(iterations, int):
            raise TypeError("The number of iterations must be an int")
        if not (isinstance(stepsize, float) or callable(stepsize)):
            raise TypeError("The stepsize must be a float or a function")

        actv_projection = False
        if len(self.agent.problem.constraints) != 0:
            actv_projection = True

        if self.enable_log:
            dims = [iterations]
            for dim in self.x.shape:
                dims.append(dim)
            self.sequence = np.zeros(dims)

        for k in range(iterations):
            if not isinstance(stepsize, float):
                step = stepsize(k)
            else:
                step = stepsize
            self.iterate_run(stepsize=step, projection=actv_projection)

            if self.enable_log:
                self.sequence[k] = self.x
            
            if verbose:
                if self.agent.id == 0:
                    print('Iteration {}'.format(k), end="\r")

        if self.enable_log:
            return self.sequence


class BlockSubgradientMethod(BlockConsensus):
    """Distributed block subgradient method. This is a special instance of the Block Proximal Method in [FaNo19]_

    At each iteration, the agent can update its local estimate or not at each iteration according to a certain probability (awakening_probability).
    From the perspective of agent :math:`i` the algorithm works as follows. At iteration :math:`k` if the agent is awake, it selects a random block :math:`\\ell_i^k` of its local solution and updates

    .. math::

        \\begin{align}
        y_i^k &= \\sum_{j\\in\\mathcal{N}_i} w_{ij} x_{j\\mid i}^k \\\\
        x_{i,\\ell}^{k+1} &= \\begin{cases}
             \\Pi_{X_\\ell}\\left[y_i^k - \\alpha_i^k [\\tilde{\\nabla} f_i(y_i^k)]_{\\ell}\\right] & \\text{if } \\ell = \\ell_i^k \\\\
             x_{i,\\ell}^{k} & \\text{otherwise}
             \\end{cases}
        \\end{align}

    then it broadcasts :math:`x_{i,\\ell_i^k}^{k+1}` to its out-neighbors. Otherwise (if the agent is not awake) :math:`x_{i}^{k+1}=x_i^k`.
    Here :math:`\\mathcal{N}_i` is the current set of in-neighbors and :math:`x_{j\\mid i},j\\in\\mathcal{N}_i` is the local copy of :math:`x_j` available at node :math:`i` and :math:`x_{i,\\ell}` denotes the :math:`\\ell`-th block of :math:`x_i`. 
    The weight matrix :math:`W=[w_{ij}]_{i,j=1}^N` should be doubly stochastic.

    Notice that if there is only one block and awakening_probability=1 the :class:`BlockSubgradientMethod` reduces to the :class:`SubgradientMethod`.
    """

    def __init__(self, agent: Agent, initial_condition: np.ndarray, **kwargs):
        if not isinstance(agent, Agent):
            raise TypeError("Agent must be an Agent")
        super(BlockSubgradientMethod, self).__init__(agent, initial_condition, **kwargs)

    def _update_local_solution(self, x: np.ndarray, selected_block: Tuple, stepsize: float = 0.1, projection: bool = False):
        block = list(selected_block)
        y = x
        # Perform a sugradient step
        block_method = getattr(self.agent.problem.objective_function, "blocksubgradient", None)

        if callable(block_method):
            self.x[block] = y[block] - stepsize * self.agent.problem.objective_function.blocksubgradient(y, block)[block]
        else:
            self.x[block] = y[block] - stepsize * self.agent.problem.objective_function.subgradient(y)[block]

        if isinstance(self.agent.problem.objective_function, SubmodularFn):
            for i in np.arange(len(block)):
                if self.x[block[i]] > 1:
                    self.x[block[i]] = 1
                if self.x[block[i]] < 0:
                    self.x[block[i]] = 0
        if projection:
            # TODO: single block projection with multiple sets
            old_x = deepcopy(self.x)
            self.x = self.agent.problem.project_on_constraint_set(self.x)
            for idx, value in enumerate(self.x):
                if idx not in block:
                    if old_x[idx] != value:
                        warnings.warn("Constraints are not separable on blocks. Convergence is not guaranteed.")
                        break

    def run(self, iterations: int = 1000, stepsize: Union[float, Callable] = 0.1, verbose: bool=False) -> np.ndarray:
        """Run the algorithm for a given number of iterations

        Args:
            iterations: Number of iterations. Defaults to 1000.
            stepsize: If a float is given as input, the stepsize is constant. 
                                                         If a function is given, it must take an iteration k as input and output the corresponding stepsize.. Defaults to 0.1.
            verbose: If True print some information during the evolution of the algorithm. Defaults to False.

        Raises:
            TypeError: The number of iterations must be an int
            TypeError: The stepsize must be a float or a callable
            ValueError: Only sets (children of AstractSet) with explicit projections are currently supported
            ValueError: Only one constraint per time is currently supported

        Returns:
            return the sequence of estimates if enable_log=True.
        """
        if not isinstance(iterations, int):
            raise TypeError("The number of iterations must be an int")
        if not (isinstance(stepsize, float) or callable(stepsize)):
            raise TypeError("The stepsize must be a float or a function")

        actv_projection = False
        if len(self.agent.problem.constraints) != 0:
            actv_projection = True

        if self.enable_log:
            dims = [iterations]
            for dim in self.x.shape:
                dims.append(dim)
            self.sequence = np.zeros(dims)

        for k in range(iterations):
            if not isinstance(stepsize, float):
                step = stepsize(k)
            else:
                step = stepsize
            self.iterate_run(stepsize=step, projection=actv_projection)

            if self.enable_log:
                self.sequence[k] = self.x
            
            if verbose:
                if self.agent.id == 0:
                    print('Iteration {}'.format(k), end="\r")

        if self.enable_log:
            return self.sequence





import numpy as np
import autograd.numpy as anp
import random
import warnings
from typing import Union
from .abstract_function import AbstractFunction
from .utilities import check_input


class Abs(AbstractFunction):
    """Absolute value (element-wise)

    .. math::

        f(x)=|x|

    with :math:`x: \\mathbb{R}^{n}`.

    Args:
        fn (AbstractFunction): input function

    Raises:
        TypeError: input must be a function object
        NotImplementedError: only 1, 2 and inf norms are currently supported
    """

    def __init__(self, fn: AbstractFunction):
        if not isinstance(fn, AbstractFunction):
            raise TypeError("Input must be a AbstractFunction object")
        self.fn = fn
        # if not fn.is_differentiable:
        #     warnings.warn(
        #         'Composition with a nondifferentiable function will lead to an\
        #              error when asking for a subgradient')

        # if not fn.is_affine:
        #     warnings.warn(
        #         'Composition with a non affine function will lead to an error \
        #             when asking for a subgradient')

        self.input_shape = fn.input_shape
        self.output_shape = fn.output_shape

        self.differentiable = False
        self.affine = False
        self.quadratic = False

        super().__init__()

    def _expression(self):
        expression = 'Abs({})'.format(self.fn._expression())
        return expression

    def _to_cvxpy(self):
        import cvxpy as cvx
        return cvx.abs(self.fn._to_cvxpy())

    def _extend_variable(self, n_var, axis, pos):
        return Abs(self.fn._extend_variable(n_var, axis, pos))

    @check_input
    def eval(self, x: np.ndarray) -> np.ndarray:
        return anp.abs(self.fn.eval(x)).reshape(self.output_shape)



import numpy as np
import autograd.numpy as anp
import random
import warnings
from typing import Union
from .abstract_function import AbstractFunction
from .utilities import check_input


class Norm(AbstractFunction):
    """Norm of a function (supporte norms are 1, 2, inf)

    .. math::

        f(x)=\\|x\\|

    with :math:`x: \\mathbb{R}^{n}`.

    Args:
        fn (AbstractFunction): input function
        order (int, optional): order of the norm. Can be 1, 2 or np.inf. Defaults to 2.

    Raises:
        TypeError: input must be a function object
        NotImplementedError: only 1, 2 and inf norms are currently supported
    """

    def __init__(self, fn: AbstractFunction, order: Union[int, float] = None, axis: int = None):
        if not isinstance(fn, AbstractFunction):
            raise TypeError("Input must be a AbstractFunction object")
        self.fn = fn
        if not fn.is_differentiable:
            warnings.warn(
                'Composition with a nondifferentiable function will lead to an\
                     error when asking for a subgradient')

        if not fn.is_affine:
            warnings.warn(
                'Composition with a non affine function will lead to an error \
                    when asking for a subgradient')

        if order not in [1, 2, np.inf, None]:
            raise NotImplementedError
        self.order = order
        self.axis = axis
        self.input_shape = fn.input_shape
        self.last_input_shape = fn.output_shape
        self.output_shape = (1, 1)

        self.differentiable = False
        self.affine = False
        self.quadratic = False

        super().__init__()

    def _expression(self):
        expression = 'Norm({}, order={})'.format(self.fn._expression(), self.order)
        return expression
    
    def _to_cvxpy(self):
        import cvxpy as cvx
        if self.order == 2 or self.order == None:
            return cvx.norm(self.fn._to_cvxpy())
        
        if self.order == 1:
            return cvx.norm1(self.fn._to_cvxpy())

        if self.order == np.inf:
            return cvx.norm_inf(self.fn._to_cvxpy())
    
    def _extend_variable(self, n_var, axis, pos):
        return Norm(self.fn._extend_variable(n_var, axis, pos), self.order, self.axis)

    @check_input
    def eval(self, x: np.ndarray) -> np.ndarray:
        return anp.linalg.norm(self.fn.eval(x), ord=self.order, axis=self.axis).reshape(self.output_shape)

    def _alternative_jacobian(self, x: np.ndarray, **kwargs) -> np.ndarray:
        if self.order in [1, 2, np.inf, None]:
            if not self.fn.is_differentiable:
                warnings.warn("Composition of non affine functions. The Jacobian may be not correct.")
            if self.last_input_shape[1] != 1:
                raise NotImplementedError
            p_subg = np.zeros(self.last_input_shape)
            pt = self.fn.eval(x)
            if self.order == 1:
                for i in range(self.last_input_shape[0]):
                    if pt[i] != 0:
                        p_subg[i] = np.sign(pt[i])
                    else:
                        p_subg[i] = random.uniform(-1, 1)
            if self.order == 2 or self.order == None:
                for i in range(self.last_input_shape[0]):
                    if pt[i] != 0:
                        p_subg[i] = pt[i] / np.linalg.norm(pt, ord=2)
                    else:
                        p_subg[i] = random.uniform(-1, 1)
            if self.order == np.inf:
                n1 = np.linalg.norm(pt, ord=1, axis=1).reshape(self.input_shape)
                idx = np.argmax(n1, axis=1)
                w1 = np.random.rand(self.input_shape)
                w2 = np.zeros(self.input_shape)
                w2 += w1
                w2[idx] = 0
                w = w1 - w2
                w = w / sum(w)
                for i in idx:
                    if pt[i] != 0:
                        p_subg[i] = w[i] * np.sign(pt[i])
                    else:
                        p_subg[i] = w[i] * random.uniform(-1, 1)
            subg = p_subg.transpose() @ self.fn.jacobian(x, **kwargs)
            return subg
        else:
            raise NotImplementedError



from typing import Union, Any
from threading import Event
from ..communicators import Communicator, MPICommunicator
from ..problems import Problem


class Agent():
    """The Agent object represents an agent in a network with communication capabilities

        Args:
            in_neighbors (list): list of agents from which communication is received
            out_neighbors (list): list of agents to which information is send
            communicator (Communicator, optional): a Communicator object used to perform communications (if none is provided, it is automatically set to MPICommunicator). Defaults to None.
            in_weights (list or dict, optional): list or dict containing weights to assign to information coming from each in-neighbor. If a list is provided, it must have lenght equal to the number of agents in the network. If a dict is provided, it must have a key for each in-neighbor and, associated to it, the correspondig weight. Defaults to None, implies equal in_weights to in-neighbors.
            out_weights (list or dict, optional): list or dict containing weights to assign to out-neighbor. If a list is provided, it must have lenght equal to the number of agents in the network. If a dict is provided, it must have a key for each out-neighbor and, associated to it, the correspondig weight. Defaults to None, implies equal in_weights to out-neighbors.
            auto_local (bool, optional): If False the (in-)weight for the local agent must be provided. Otherwise it is set automatically, provided that the in_weights have sum in [0,1]. Defaults to True.

        Attributes:
            id (int): id of the Agent
            in_neighbors (list): list of in-neighbors
            out_neighbors (list): list of out-neighbors
            in_weights (dict): a dict containing weights to assign to information coming from each in-neighbor.
            out_weights (dict): a dict containing weights to assign to out-neighbors.
            communicator (Communicator): Communicator object used to perform communications.
            problem (Problem): Local optimization problem.
        """

    def __init__(self, in_neighbors: list = None, out_neighbors: list = None, communicator: Communicator = None,
                 in_weights: Union[list, dict] = None, out_weights: Union[list, dict] = None, auto_local: bool = True):

        if communicator is not None:
            if isinstance(communicator, Communicator):
                self.communicator = communicator
            else:
                raise ValueError(
                    "the communicator must be and instance of the Communicator class.")
        else:
            self.communicator = MPICommunicator()

        self.id = self.communicator.rank

        if in_neighbors is None:
            in_neighbors = list(range(self.communicator.size)).pop(self.id)

        if out_neighbors is None:
            out_neighbors = in_neighbors

        # Set neighbors
        self.in_neighbors = None
        self.out_neighbors = None
        self.set_neighbors(in_neighbors, out_neighbors)

        self.in_weights = {}
        self.set_weights(in_weights, out_weights, auto_local)

        self.problem = None

    def set_neighbors(self, in_neighbors: list, out_neighbors: list):
        """Set in and out neighbors

        Args:
            in_neighbors: list of agents from which communication is received
            out_neighbors: list of agents to which information is send
        """

        if isinstance(in_neighbors, list) and isinstance(out_neighbors, list):
            self.in_neighbors = in_neighbors
            self.out_neighbors = out_neighbors

            # Remove self-loops if any
            if self.id in self.in_neighbors:
                self.in_neighbors.pop(self.id)
            if self.id in self.out_neighbors:
                self.out_neighbors.pop(self.id)

    def set_weights(self, in_weights: Union[list, dict] = None, out_weights: Union[list, dict] = None, auto_local: bool = True):
        """Set in_weights to assign to in-neighbors and the one for agent itself.

        Args:
            in_weights: list or dict contatining in_weights to assign to information coming from each in-neighbor. If a list is provided, it must have lenght equal to the number of agents in the network. If a dict is provided, it must have a key for each in-neighbor and, associated to it, the correspondig weight. Defaults to None, implies equal in_weights to in-neighbors.
            out_weights: list or dict contatining in_weights to assign to out-neighbors. If a list is provided, it must have lenght equal to the number of agents in the network. If a dict is provided, it must have a key for each out-neighbor and, associated to it, the correspondig weight. Defaults to None, implies equal in_weights to out neighbors.
            auto_local: If False the weight for the local agent must be provided. Otherwise it is set automatically, provided that the in_weights have sum in [0,1]. Defaults to True.
        Raises:
            ValueError: If a dict is provided as argument, it must contain a key for each in-neighbor.
            ValueError: Input must be list or dict
            ValueError: If auto_local is not False, the provided in_weights must have sum in [0,1]
        """
        self.in_weights = {}
        self.out_weights = {}
        if in_weights is not None:
            # if a list is provided with length equal to the number of agents in the network
            # it is converted in a dict
            if isinstance(in_weights, list) and len(in_weights) == self.communicator.size:
                for neighbor in self.in_neighbors:
                    self.in_weights[neighbor] = in_weights[neighbor]

            # if a dict is provided it is checked if all in neighbors have a weight
            else:
                if isinstance(in_weights, dict):
                    for neighbor in self.in_neighbors:
                        if neighbor in in_weights:
                            self.in_weights[neighbor] = in_weights[neighbor]
                        else:
                            raise ValueError(
                                "A weight for each in-neighbor must be provided.")
                else:
                    raise ValueError("The in_weights argument must be a list or a dict. \
                            If a list is provided, it must have lenght equal to the number of agents in the network. \
                            If a dict is provided, it must have a key for each in-neighbor and, associated to it, the correspondig weight.")

            # Assign self weight (if sum of provided in_weights in [0,1])
            if not auto_local:
                self.in_weights[self.id] = in_weights[self.id]
            else:
                if 0 <= sum(self.in_weights.values()) <= 1:
                    self.in_weights[self.id] = 1 - sum(self.in_weights.values())
                else:
                    raise ValueError(
                        "If auto_local is set to True, the provided in_weights must have sum in [0,1]")
        else:
            for neighbor in self.in_neighbors:
                self.in_weights[neighbor] = 1/len(self.in_neighbors)

        if out_weights is not None:
            # if a list is provided with length equal to the number of agents in the network
            # it is converted in a dict
            if isinstance(out_weights, list) and len(out_weights) == self.communicator.size:
                for neighbor in self.out_neighbors:
                    self.out_weights[neighbor] = out_weights[neighbor]

            # if a dict is provided it is checked if all in neighbors have a weight
            else:
                if isinstance(out_weights, dict):
                    for neighbor in self.out_neighbors:
                        if neighbor in out_weights:
                            self.out_weights[neighbor] = out_weights[neighbor]
                        else:
                            raise ValueError(
                                "A weight for each in-neighbor must be provided.")
                else:
                    raise ValueError("The in_weights argument must be a list or a dict. \
                            If a list is provided, it must have lenght equal to the number of agents in the network. \
                            If a dict is provided, it must have a key for each in-neighbor and, associated to it, the correspondig weight.")
        
            # Assign self out_weight (if sum of provided in_weights in [0,1])
            if not auto_local:
                self.out_weights[self.id] = out_weights[self.id]
            else:
                if 0 <= sum(self.out_weights.values()) <= 1:
                    self.out_weights[self.id] = 1 - sum(self.out_weights.values())
                else:
                    raise ValueError(
                        "If auto_local is set to True, the provided out_weights must have sum in [0,1]")
        else:
            for neighbor in self.out_neighbors:
                self.out_weights[neighbor] = 1/len(self.out_neighbors)

    def neighbors_exchange(self, obj: Any, dict_neigh=False, stop_event: Event=None):
        """Exchange data with neighbors (synchronously). Send obj to the out-neighbors and receive received_obj from in-neighbors

        Args:
            obj: object to send
            dict_neigh: True if obj contains a dictionary with different objects for each neighbor. Defaults to False.

        Returns:
            dict: a dictionary containing an object for each in-neighbor
        """
        received_obj = self.communicator.neighbors_exchange(
            obj, self.in_neighbors, self.out_neighbors, dict_neigh, stop_event)
        return received_obj

    def neighbors_send(self, obj: Any):
        """Send data to out-neighbors

        Args:
            obj: object to send
        """
        self.communicator.neighbors_send(obj, self.out_neighbors)

    def neighbors_receive_asynchronous(self):
        """Receive data from in-neighbors (if any have been sent)

        Returns:
            dict: a dictionary containing an object for each in-neighbor that has sent one
        """
        received_obj = self.communicator.neighbors_receive_asynchronous(
            self.out_neighbors)
        return received_obj

    def set_problem(self, problem: Problem):
        """set the local optimization problem

        Args:
            problem (Problem): Problem object

        Raises:
            TypeError: Input must be a Problem object
        """
        if not isinstance(problem, Problem):
            raise TypeError("Input must be a Problem object")
        self.problem = problem




import numpy as np
import autograd.numpy as anp
import warnings
from typing import Union
from .abstract_function import AbstractFunction
from .utilities import check_input
from .norm import Norm


class SquaredNorm(AbstractFunction):
    """Squared norm (supporte norms are 1, 2, inf)
     
    .. math::

        f(x)=\\|x\\|^2

    with :math:`x: \\mathbb{R}^{n}`.

    Args:
        fn (AbstractFunction): input function
        order (int, optional): order of the norm. Can be 1, 2 or np.inf. Defaults to 2.
   
    Raises:
        TypeError: input must be a function object
        NotImplementedError: only 1, 2 and inf norms are currently supported
    """

    def __init__(self, fn: AbstractFunction, order: Union[int, float]=None, axis: int=None):
        if not isinstance(fn, AbstractFunction):
            raise TypeError("Input must be a AbstractFunction object")

        # if not fn.is_differentiable:
        #     warnings.warn(
        #         'Composition with a nondifferentiable function will lead to an\
        #             error when asking for a subgradient')

        # if not fn.is_affine:
        #     warnings.warn(
        #         'Composition with a non affine function will lead to an error \
        #             when asking for a subgradient')

        self.fn = fn
        if order not in [1, 2, np.inf, None]:
            raise NotImplementedError
        if order == 2:
            self.differentiable = True

        self.order = order
        self.axis = axis
        self.input_shape = fn.input_shape
        self.output_shape = (1, 1)

        self.affine = False
        self.quadratic = False # TODO

        super().__init__()

    def _expression(self):
        expression = 'SquaredNorm({})'.format(self.fn._expression())
        return expression 

    def _to_cvxpy(self):
        import cvxpy as cvx
        return cvx.square(cvx.norm(self.fn._to_cvxpy()))

    @check_input
    def eval(self, x: np.ndarray) -> np.ndarray:
        # TODO: order
        return anp.power(anp.linalg.norm(self.fn.eval(x), ord=self.order, axis=self.axis), 2).reshape(self.output_shape)
    
    def _extend_variable(self, n_var, axis, pos):
        return SquaredNorm(self.fn._extend_variable(n_var, axis, pos), self.order, self.axis)

    @check_input
    def _alternative_jacobian(self, x: np.ndarray, **kwargs) -> np.ndarray:
        if self.order in [1, 2, np.inf, None]:
            if not self.fn.is_differentiable:
                warnings.warn("Composition of nondifferentiable functions. The Jacobian may be not correct.")
                # raise ValueError(
                #     "Composition of nondifferentiable functions. \
                #         No subgradients available.")
            if not self.fn.is_affine:
                warnings.warn("Composition of non affine functions. The Jacobian may be not correct.")
                # raise ValueError(
                #     "Composition of non affine function. \
                #         No subgradients available.")
            if self.order in [2, None]: 
                return 2 * self.fn.eval(x).transpose() @ self.fn.jacobian(x, **kwargs)
            else:
                norm = Norm(self.fn, order=self.order)
                return 2 * norm.jacobian(x, **kwargs) * \
                    np.linalg.norm(self.fn.eval(x), ord=self.order)
        else:
            raise NotImplementedError
    
